/***************************************************************************
 *   Copyright (C) 2007 by David Bitseff                                   *
 *   dbitsef@zipcon.net                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <QDockWidget>
#include <QFileInfo>
#include <QFileDialog>

#include "mainviewer.h"

//
// This has gotten wierd and complicated.  The point of the timer
// is to have the viewerResized() signal emitted after the mouse button
// is resleased.  This is used only when the parent widget is a dock
// widget.
//
// The MainViewer is used as a dock widget, and as a window.  The '[' and
// ']' keys scale the image.  The 'o' key shows the image in original size.
// The 'm' key resizes to the previous 'o' size.  The 'q' key closes the
// viewer.
//
// I don't remember why there is a graphicsscene being used instead
// of a label for the pixmap.
//
MainViewer::MainViewer ( QWidget* parent ) : QWidget(parent)
{
	setupUi ( this );

	m_pitem = m_scene.addPixmap(QPixmap(0,0));
	m_graphicsView->setScene(&m_scene);
	m_scene.setBackgroundBrush( Qt::black );
	m_resizedself = false;
	m_statusWidget->setVisible(false);
	timer = new QTimer(this);
	timer->setInterval(1000);
	timer->setSingleShot(true);
	connect(timer, SIGNAL(timeout()), this, SLOT(checkResized()));
}


QSize MainViewer::getViewerSize()
{
	return m_graphicsView->maximumViewportSize();
}

// if parent() is a dockwidget, then emit viewerResized() using the timer,
// otherwise emit a signal if user changes the widget size
void MainViewer::resizeEvent(QResizeEvent* e)
{
	static QSize lastSize(0,0);
	if (m_resizedself)
		m_resizedself = false;
	else if (lastSize != getViewerSize())
	{
		m_scaled_size = last_M = lastSize = getViewerSize();
		if (parent() && parent()->inherits("QDockWidget")
				  && QApplication::mouseButtons() != Qt::NoButton)
			timer->start();
		else
			emit viewerResized( getViewerSize() );
	}
}

void MainViewer::keyPressEvent(QKeyEvent* e)
{
	if (e->key() == Qt::Key_Escape)
	{
		if (parent() && parent()->inherits("QDockWidget"))
			qobject_cast<QDockWidget*>(parent())->close();
		else
			close();
	}
	// scale
	else if (e->key() == Qt::Key_F7 || e->key() == Qt::Key_F8)
	{
		if (e->key() == Qt::Key_F7)
			m_scaled_size *= 0.9;
		else if (e->key() == Qt::Key_F8)
 			m_scaled_size *= 1.1;

		m_pitem->setPixmap( m_pix.scaled(m_scaled_size,
										Qt::KeepAspectRatio,
										Qt::SmoothTransformation) );
		rescaleViewer();
	}
	// old size toggle
	else if (e->key() == Qt::Key_F5)
	{
		if (m_pitem->pixmap().size() == m_pix.size())
		{
			m_pitem->setPixmap( m_pix.scaled(last_M,
										Qt::KeepAspectRatio,
										Qt::SmoothTransformation) );
		}
		else
		{
			last_M = m_pitem->pixmap().size();
			m_pitem->setPixmap( m_pix );
		}
		rescaleViewer();
	}
	// total size of image
	else if (e->key() == Qt::Key_F6)
	{
		if (m_pitem->pixmap().size() != m_pix.size())
		{
			m_scaled_size = last_M = m_pitem->pixmap().size();
			m_pitem->setPixmap(m_pix);
			rescaleViewer();
		}
	}
	// save to file
	else if (e->key() == Qt::Key_F9)
	{
		QString fileName = QFileDialog::getSaveFileName(this,
				"Save an image", "untitled.png");

		if (!fileName.isEmpty()
				   && QFileInfo(QFileInfo(fileName).path()).isWritable())
		{
			m_pitem->pixmap().save(fileName, "png");
			QString msg(QFileInfo(fileName).fileName());
			msg += " saved";
			setStatusMessage(msg);
		}
	}
}

void MainViewer::setPixmap(QPixmap& p, bool resized)
{
	setPixmap(QPixmap(p), resized);
}

// the resized boolean tells the viewer to rescale either itself
// or the pixmap.  scale the viewer by default
void MainViewer::setPixmap(const QPixmap& p, bool resized)
{
	m_pix = p;
	m_orig_size = m_pix.size();
	m_scaled_size = getViewerSize();
	last_M = m_scaled_size;

	if (!(p.isNull() || getViewerSize() == p.size()))
	{
		if (resized)
		{
			m_pitem->setPixmap( p );
			rescaleViewer();
		}
		else
			rescalePixmap();
	}
	else
	{
		m_pitem->setPixmap( p );
		QSize size = p.size();
		m_graphicsView->setSceneRect(0, 0, size.width(), size.height());
	}
}

void MainViewer::setStatusMessage(QString& s)
{
	m_statusWidget->setText( s );
}

void MainViewer::showEvent(QShowEvent* e)
{
	if (!e->spontaneous())
		emit viewerResized( getViewerSize() );
}

QPixmap MainViewer::pixmap()
{
	return m_pix;
}

// used by dockwidget children, look for no buttons, otherwise check later
void MainViewer::checkResized()
{
	if(QApplication::mouseButtons() == Qt::NoButton)
		emit viewerResized( getViewerSize() );
	else
		timer->start();
}

MainViewer::~MainViewer()
{
	disconnect(timer);
	delete timer;
}

// rescale the pixmap to fit the viewer
void MainViewer::rescalePixmap()
{
	m_pitem->setPixmap( m_pix.scaled(getViewerSize(),
									Qt::KeepAspectRatio,
									Qt::SmoothTransformation) );
	QSize size = m_pitem->pixmap().size();
	m_graphicsView->setSceneRect(0, 0, size.width(), size.height());
}

// rescale the viewer to fit the pixmap
void MainViewer::rescaleViewer()
{
	m_resizedself = true;
	QSize px_size = m_pitem->pixmap().size();
	QSize v_size  = m_graphicsView->maximumViewportSize();
	QSize ds = px_size - v_size;
	const QSize fs = frameSize() + ds;

	if (parent() && parent()->inherits("QDockWidget"))
		qobject_cast<QDockWidget*>(parent())->resize(fs);
	else
		resize(fs);

	m_graphicsView->setSceneRect(0, 0, px_size.width(), px_size.height());
}

void MainViewer::hideEvent(QHideEvent* e)
{
	if (!e->spontaneous())
		emit viewerHidden();
}


