/***************************************************************************
 *   Copyright (C) 2007 by David Bitseff                                   *
 *   dbitsef@zipcon.net                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <QDockWidget>
#include <QToolBar>
#include <QSettings>
#include <QFileDialog>
#include <QMessageBox>


#include "qosmic.h"
#include "mainwindow.h"
#include "genomevector.h"
#include "flam3util.h"
#include "mainviewer.h"
#include "paletteeditor.h"
#include "colorselector.h"
#include "mutationwidget.h"
#include "camerasettingswidget.h"
#include "colorsettingswidget.h"
#include "edittrianglewidget.h"
#include "colorbalancewidget.h"
#include "imgsettingswidget.h"
#include "trianglecoordswidget.h"
#include "triangledensitywidget.h"
#include "mainpreviewwidget.h"
#include "variationswidget.h"
#include "directoryviewwidget.h"
#include "statuswidget.h"
#include "scripteditwidget.h"
#include "selectgenomewidget.h"
#include "viewerpresetswidget.h"
#include "selecttrianglewidget.h"
#include "coordinatemark.h"


MainWindow::MainWindow() : QMainWindow()
{
	setupUi(this);

	f.ngenomes = 0;
	f.genomes = genomes.data();
	genomes.setSelectedIndex(0);
	lastSelected = 0;
	m_fileViewer = 0;
	m_dialogsEnabled = true;

	// These are the request instances sent to the renderthread
	m_preview_request.setGenome(genomes.data());
	m_preview_request.setName("preview");
	m_preview_request.setType(RenderRequest::Preview);

	m_viewer_request.setGenome(genomes.data());
	m_viewer_request.setName("viewer");
	m_viewer_request.setType(RenderRequest::Image);

	m_file_request.setGenome(genomes.data());
	m_file_request.setName("file.png");
	m_file_request.setType(RenderRequest::File);

	// figure editor
	logInfo("MainWindow::MainWindow : creating FigureEditor");
	m_xfeditor = new FigureEditor(&genomes, this);
	m_xfeditview->setScene(m_xfeditor);
	connect(m_xfeditor, SIGNAL(triangleModifiedSignal(Triangle*)),
			this, SLOT(triangleModifiedSlot(Triangle*)));
	connect(m_xfeditor, SIGNAL(triangleSelectedSignal(Triangle*)),
			this, SLOT(triangleSelectedSlot(Triangle*)));
	connect(m_xfeditor, SIGNAL(coordinateChangeSignal(double,double)),
			this, SLOT(updateStatus(double,double)));


	QDockWidget *dock;
	QDockWidget *lastDock;
	QList<QAction*> dockActions;
	setDockOptions( // AnimatedDocks |
			AllowTabbedDocks
			| AllowNestedDocks | VerticalTabs);
	setCorner(Qt::BottomLeftCorner,  Qt::LeftDockWidgetArea);
	setCorner(Qt::BottomRightCorner, Qt::RightDockWidgetArea);

	// main preview widget
	logInfo("MainWindow::MainWindow : creating PreviewWidget");
	dock = new QDockWidget(tr("Preview"), this);
	dock->setObjectName(dock->windowTitle());
	dock->setAllowedAreas(Qt::AllDockWidgetAreas);
	m_previewWidget = new MainPreviewWidget(&genomes, dock);
	dock->setWidget(m_previewWidget);
	addDockWidget(Qt::RightDockWidgetArea, dock);
	dockActions << dock->toggleViewAction();
	lastDock = dock;
	connect(m_previewWidget, SIGNAL(previewMoved()), this, SLOT(render()));

	// select a triangle widget
	logInfo("MainWindow::MainWindow : creating SelectTriangleWidget");
	dock = new QDockWidget(tr("Triangles"), this);
	dock->setObjectName(dock->windowTitle());
	dock->setAllowedAreas(Qt::AllDockWidgetAreas);
	m_selectTriangleWidget = new SelectTriangleWidget(&genomes, dock);
	dock->setWidget(m_selectTriangleWidget);
	splitDockWidget(lastDock, dock, Qt::Vertical);
	dockActions << dock->toggleViewAction();
	lastDock = dock;
	connect(m_xfeditor, SIGNAL(triangleSelectedSignal(Triangle*)),
			m_selectTriangleWidget, SLOT(triangleSelectedSlot(Triangle*)));
	connect(m_selectTriangleWidget, SIGNAL(dataChanged()), this, SLOT(render()));


	// triangle density widget
	logInfo("MainWindow::MainWindow : creating TriangleDensityWidget");
	dock = new QDockWidget(tr("Densities"), this);
	dock->setObjectName(dock->windowTitle());
	dock->setAllowedAreas(Qt::AllDockWidgetAreas);
	m_triangleDensityWidget = new TriangleDensityWidget(&genomes, m_xfeditor, dock);
	dock->setWidget(m_triangleDensityWidget);
	splitDockWidget(lastDock, dock, Qt::Vertical);
	dockActions << dock->toggleViewAction();
	lastDock = dock;
	connect(m_xfeditor, SIGNAL(triangleSelectedSignal(Triangle*)),
		m_triangleDensityWidget, SLOT(triangleSelectedSlot(Triangle*)));
	connect(m_triangleDensityWidget, SIGNAL(dataChanged()),
		m_selectTriangleWidget, SLOT(reset()));
	connect(m_selectTriangleWidget, SIGNAL(dataChanged()),
		m_triangleDensityWidget, SLOT(reset()));
	connect(m_triangleDensityWidget, SIGNAL(dataChanged()),
		this, SLOT(render()));
	connect(m_triangleDensityWidget, SIGNAL(genomeMerged()),
			this, SLOT(scriptFinishedSlot()));

	// main viewer
	logInfo("MainWindow::MainWindow : creating ViewerWidget");
	dock = new QDockWidget(tr("Viewer"), this);
	dock->setObjectName(dock->windowTitle());
	dock->setAllowedAreas(Qt::AllDockWidgetAreas);
	m_viewer = new MainViewer(dock);
	dock->setWidget(m_viewer);
	addDockWidget(Qt::LeftDockWidgetArea, dock);
	dockActions << dock->toggleViewAction();
	dock->hide();
	dock->setFloating(true);

	// status widget
	logInfo("MainWindow::MainWindow : creating StatusWidget");
	m_statusWidget = new StatusWidget(this);

	// image settings widget
	logInfo("MainWindow::MainWindow : creating ImageSettingsWidget");
	dock = new QDockWidget(tr("Image"), this);
	dock->setObjectName(dock->windowTitle());
	dock->setAllowedAreas(Qt::AllDockWidgetAreas);
	m_imageSettingsWidget = new ImageSettingsWidget(&genomes, dock);
	dock->setWidget(m_imageSettingsWidget);
	splitDockWidget(lastDock, dock, Qt::Vertical);
	dockActions << dock->toggleViewAction();
	connect(m_imageSettingsWidget, SIGNAL(dataChanged()), this, SLOT(render()));
	connect(m_imageSettingsWidget, SIGNAL(symmetryAdded()),
			this, SLOT(scriptFinishedSlot()));
	lastDock = dock;


	// camera settings widget
	logInfo("MainWindow::MainWindow : creating CameraSettingsWidget");
	dock = new QDockWidget(tr("Camera"), this);
	dock->setObjectName(dock->windowTitle());
	dock->setAllowedAreas(Qt::AllDockWidgetAreas);
	m_cameraSettingsWidget = new CameraSettingsWidget(&genomes, dock);
	dock->setWidget(m_cameraSettingsWidget);
	tabifyDockWidget(lastDock, dock);
	dockActions << dock->toggleViewAction();
	connect(m_cameraSettingsWidget, SIGNAL(dataChanged()), this, SLOT(render()));
	connect(m_previewWidget, SIGNAL(previewMoved()),
			m_cameraSettingsWidget, SLOT(updateFormData()));
	lastDock = dock;


	// triangle coords widget
	logInfo("MainWindow::MainWindow : creating TriangleCoordsWidget");
	dock = new QDockWidget(tr("Coordinates"), this);
	dock->setObjectName(dock->windowTitle());
	dock->setAllowedAreas(Qt::AllDockWidgetAreas);
	m_coordsWidget = new TriangleCoordsWidget(&genomes, dock);
	dock->setWidget(m_coordsWidget);
	tabifyDockWidget(lastDock, dock);
	dockActions << dock->toggleViewAction();
	lastDock = dock;
	connect(m_xfeditor, SIGNAL(triangleSelectedSignal(Triangle*)),
			 m_coordsWidget, SLOT(triangleSelectedSlot(Triangle*)));
	connect(m_xfeditor, SIGNAL(triangleModifiedSignal(Triangle*)),
			m_coordsWidget, SLOT(triangleModifiedSlot(Triangle*)));
	connect(m_coordsWidget, SIGNAL(dataChanged()), this, SLOT(render()));

	// triangle editing widget
	logInfo("MainWindow::MainWindow : creating EditTriangleWidget");
	dock = new QDockWidget(tr("Edit Triangle"), this);
	dock->setObjectName(dock->windowTitle());
	dock->setAllowedAreas(Qt::AllDockWidgetAreas);
	m_editTriangleWidget = new EditTriangleWidget(&genomes, m_xfeditor, dock);
	dock->setWidget(m_editTriangleWidget);
	tabifyDockWidget(lastDock, dock);
	dockActions << dock->toggleViewAction();
	connect(m_xfeditor, SIGNAL(triangleSelectedSignal(Triangle*)),
			m_editTriangleWidget, SLOT(triangleSelectedSlot(Triangle*)));
	connect(m_editTriangleWidget, SIGNAL(dataChanged()), this, SLOT(render()));


	// color settings widget
	logInfo("MainWindow::MainWindow : creating ColorSettingsWidget");
	dock = new QDockWidget(tr("Colors"), this);
	dock->setObjectName(dock->windowTitle());
	dock->setAllowedAreas(Qt::AllDockWidgetAreas);
	m_colorSettingsWidget = new ColorSettingsWidget(&genomes, dock);
	dock->setWidget(m_colorSettingsWidget);
	tabifyDockWidget(lastDock, dock);
	dockActions << dock->toggleViewAction();
	lastDock = dock;
	connect(m_xfeditor, SIGNAL(triangleSelectedSignal(Triangle*)),
			m_colorSettingsWidget, SLOT(triangleSelectedSlot(Triangle*)));
	connect(m_colorSettingsWidget, SIGNAL(dataChanged()), this, SLOT(render()));
	connect(m_colorSettingsWidget, SIGNAL(colorSelected(double)),
			this, SLOT(colorSelected(double)));

	// viewer presets widget
	logInfo("MainWindow::MainWindow : creating ViewerPresetsWidget");
	dock = new QDockWidget(tr("Viewer Presets"), this);
	dock->setObjectName(dock->windowTitle());
	dock->setAllowedAreas(Qt::AllDockWidgetAreas);
	m_viewerPresetsWidget = new ViewerPresetsWidget(&genomes, dock);
	dock->setWidget(m_viewerPresetsWidget);
	tabifyDockWidget(lastDock, dock);
	dockActions << dock->toggleViewAction();
	lastDock = dock;
	connect(m_viewerPresetsWidget, SIGNAL(dataChanged()),
			this, SLOT(presetSelectedSlot()));

	// color balance widget
	logInfo("MainWindow::MainWindow : creating ColorBalanceWidget");
	dock = new QDockWidget(tr("Color Balance"), this);
	dock->setObjectName(dock->windowTitle());
	dock->setAllowedAreas(Qt::AllDockWidgetAreas);
	m_colorBalanceWidget = new ColorBalanceWidget(&genomes, dock);
	dock->setWidget(m_colorBalanceWidget);
	tabifyDockWidget(lastDock, dock);
	dockActions << dock->toggleViewAction();
	connect(m_colorBalanceWidget, SIGNAL(dataChanged()), this, SLOT(render()));
	connect(m_colorBalanceWidget, SIGNAL(paletteChanged()),
			this, SLOT(paletteHueChangedAction()));
	connect(m_colorBalanceWidget, SIGNAL(imageFormatChanged(int)),
			this, SLOT(setImageFormat(int)));
	lastDock = dock;


	// the render thread
	logInfo("MainWindow::MainWindow : creating RenderThread");
	m_rthread = new RenderThread(&f);
	m_rthread->addStatusWatcher(m_statusWidget);
	m_rthread->start();
	connect(m_rthread, SIGNAL(flameRendered(RenderEvent*)),
			this, SLOT(flameRenderedSlot(RenderEvent*)));


	// variations widget settings widget
	logInfo("MainWindow::MainWindow : creating VariationsWidget");
	dock = new QDockWidget(tr("Variations"), this);
	dock->setObjectName(dock->windowTitle());
	dock->setAllowedAreas(Qt::AllDockWidgetAreas);
	m_variationsWidget = new VariationsWidget(&genomes, dock);
	dock->setWidget(m_variationsWidget);
	tabifyDockWidget(lastDock, dock);
	dockActions << dock->toggleViewAction();
	connect(m_xfeditor, SIGNAL(triangleSelectedSignal(Triangle*)),
			m_variationsWidget, SLOT(triangleSelectedSlot(Triangle*)));
	connect(m_variationsWidget, SIGNAL(dataChanged()), this, SLOT(render()));
	lastDock = dock;

	// directory view widget
	logInfo("MainWindow::MainWindow : creating DirectoryViewWidget");
	dock = new QDockWidget(tr("Browse"), this);
	dock->setObjectName(dock->windowTitle());
	dock->setAllowedAreas(Qt::AllDockWidgetAreas);
	m_directoryViewWidget = new DirectoryViewWidget(dock);
	dock->setWidget(m_directoryViewWidget);
	addDockWidget(Qt::LeftDockWidgetArea, dock);
	dockActions << dock->toggleViewAction();
	dock->hide();
	connect(m_directoryViewWidget, SIGNAL(flam3FileSelected(const QString&,bool)),
			this, SLOT(flam3FileSelectedAction(const QString&,bool)));


	// palettes
	logInfo("MainWindow::MainWindow : creating PalettesWidget");
	dock = new QDockWidget(tr("Palettes"), this);
	dock->setObjectName(dock->windowTitle());
	dock->setAllowedAreas(Qt::AllDockWidgetAreas);
	m_paletteEditor = new PaletteEditor(dock);
	dock->setWidget(m_paletteEditor);
	splitDockWidget(lastDock, dock, Qt::Vertical);
	dockActions << dock->toggleViewAction();
	dock->hide();
	connect(m_paletteEditor, SIGNAL(paletteChanged()),
			this, SLOT(paletteChangedAction()));
	lastDock = dock;


	// mutations widget
	logInfo("MainWindow::MainWindow : creating MutationsWidget");
	dock = new QDockWidget(tr("Mutations"), this);
	dock->setObjectName(dock->windowTitle());
	dock->setAllowedAreas(Qt::AllDockWidgetAreas);
	m_mutations = new MutationWidget(&genomes, m_rthread, dock);
	dock->setWidget(m_mutations);
	tabifyDockWidget(lastDock, dock);
	dockActions << dock->toggleViewAction();
	dock->hide();
	lastDock = dock;
	connect(m_mutations, SIGNAL(genomeSelected(flam3_genome*)),
			this, SLOT(mutationSelectedSlot(flam3_genome*)));


	// script editing widget
	logInfo("MainWindow::MainWindow : creating ScriptEditorWidget");
	dock = new QDockWidget(tr("Edit Script"), this);
	dock->setObjectName(dock->windowTitle());
	dock->setAllowedAreas(Qt::AllDockWidgetAreas);
	m_scriptEditWidget = new ScriptEditWidget(this, dock);
	dock->setWidget(m_scriptEditWidget);
	tabifyDockWidget(lastDock, dock);
	dockActions << dock->toggleViewAction();
	lastDock = dock;

	// select a genome widget
	logInfo("MainWindow::MainWindow : creating SelectGenomeWidget");
	dock = new QDockWidget(tr("Genomes"), this);
	dock->setObjectName(dock->windowTitle());
	dock->setAllowedAreas(Qt::AllDockWidgetAreas);
	m_genomeSelectWidget = new SelectGenomeWidget(&genomes, m_rthread, dock);
	dock->setWidget(m_genomeSelectWidget);
	tabifyDockWidget(lastDock, dock);
	dockActions << dock->toggleViewAction();
	dock->hide();
	lastDock = dock;
	connect(m_genomeSelectWidget, SIGNAL(genomeSelected(int)),
			this, SLOT(selectGenomeSlot(int)));

	createActions();
	createMenus();
	createToolBars();
	createStatusBar();
	foreach(QAction* a, dockActions)
		settingsMenu->addAction(a);
	// install this eventfilter to capture globally the spacebar key
	qApp->installEventFilter(this);
	readSettings();
	setCurrentFile("");
}


void MainWindow::triangleSelectedSlot(Triangle* t)
{
	logFine(QString("MainWindow::triangleSelectedSlot : t=0x%1")
			.arg((long)t, 0, 16));
	selectedTriangle = t;
    lastSelected = 0;
	// sneakily update the statusBar without changing position.
	// this is just a way to avoid adding more public functions to
	// figureeditor.
	QString msg = coordsLabel.text();
	int tid = m_xfeditor->selectedTriangleIndex() + 1;
	int tnum = m_xfeditor->getNumberOfTriangles();
	int idx = msg.indexOf(']',0);
	if (idx > -1)
	{
		msg.replace(0,idx,QString(" [%1/%2").arg(tid).arg(tnum));
		coordsLabel.setText(msg);
	}
}



void MainWindow::triangleModifiedSlot(Triangle* t)
{
	logFine(QString("MainWindow::triangleModifiedSlot : t=0x%1")
			.arg((long)t, 0, 16));
	render();
}


void MainWindow::showMainViewer(QString filename)
{
	if (!m_dialogsEnabled)
		return;
	if (!m_fileViewer)
		m_fileViewer = new MainViewer();
	m_fileViewer->show();
	m_fileViewer->setPixmap(QPixmap(filename));
}

void MainWindow::flameRenderedSlot(RenderEvent* e)
{
	RenderRequest* req = e->request();

	if (req == &m_preview_request)
	{
		logFiner(QString("MainWindow::flameRenderedSlot : updating preview"));
		m_previewWidget->setPixmap(QPixmap::fromImage(req->image()));
		e->accept();
	}
	else if (req == &m_viewer_request)
	{
		logFiner(QString("MainWindow::flameRenderedSlot : updating viewer"));
		m_viewer->setPixmap(QPixmap::fromImage(req->image()));
		e->accept();
	}
	else if (req == &m_file_request)
	{
		showMainViewer(req->name());
		e->accept();
	}
}


void MainWindow::updateStatus(double posX, double posY)
{
    double tx, ty;
	QMatrix trans(m_xfeditor->transform().inverted());
	trans.map(posX, posY, &tx, &ty);
	switch (m_coordsWidget->coordType())
	{
		case TriangleCoordsWidget::Rect:
			break;
		case TriangleCoordsWidget::Degree:
		{
			Util::rectToPolarDeg(tx, ty, &tx, &ty);
			break;
		}
		case TriangleCoordsWidget::Radian:
			Util::rectToPolar(tx, ty, &tx, &ty);
	}
	CoordinateMark* mark = m_xfeditor->mark();
	if (mark->isVisible())
	{
		double mx, my;
		trans.map(mark->center().x(), mark->center().y(), &mx, &my);
		coordsLabel.setText(QString(" [%1/%2] : (%3,%4) : mark (%5,%6)")
			.arg(m_xfeditor->selectedTriangleIndex() + 1)
			.arg(m_xfeditor->getNumberOfTriangles())
			.arg(tx, 0, 'f', 4).arg(ty, 0, 'f', 4)
			.arg(mx).arg(my));
	}
	else
		coordsLabel.setText(QString(" [%1/%2] : (%3,%4)")
			.arg(m_xfeditor->selectedTriangleIndex() + 1)
			.arg(m_xfeditor->getNumberOfTriangles())
			.arg(tx, 0, 'f', 4).arg(ty, 0, 'f', 4));
}

void MainWindow::closeEvent(QCloseEvent *event)
{
	writeSettings();
	QString name(QDir::home().absoluteFilePath(TMP_FILE));
	QFile file(name);
	if (file.exists() && file.open(QIODevice::ReadWrite))
	{
		logInfo("MainWindow::closeEvent : saving genome to " + TMP_FILE);
		file.close();
		saveFile(name);
	}
	m_rthread->stopRendering();
	m_rthread->running = false;
	event->accept();
}


void MainWindow::showEvent(QShowEvent *event)
{
	logInfo("MainWindow::showEvent : showing mainwindow");
	QMainWindow::showEvent(event);

	if (event->spontaneous())
		return;

	logInfo("MainWindow::showEvent : showing widgets");
	QSettings settings;
	if (!restoreState(settings.value("state_mainwindow", QByteArray()).toByteArray()))
	{
		logWarn("MainWindow::showEvent : restoring main window to default state");
		QList<QDockWidget*> dockWidgets = findChildren<QDockWidget*>();
		foreach (QDockWidget* d, dockWidgets)
			d->hide();
	}
	else
		logFine(QString("MainWindow::showEvent : main window state restored"));

	bool flag = settings.value("visible_viewer", false).toBool();
	m_viewer->parentWidget()->setVisible(flag);

	if (!qobject_cast<QDockWidget*>(m_previewWidget->parentWidget())->isFloating())
		m_previewWidget->setPreviewMaximumSize(m_preview_request.size());

	QFile file(QDir::home().absoluteFilePath(".qosmic.lua"));
	if (file.open(QIODevice::ReadOnly))
	{
		logInfo("MainWindow::showEvent : reading lua config");
		Lua::LuaThread lua_thread(this);
		QTextStream os(&file);
		lua_thread.setLuaText(os.readAll());
		file.close();
		lua_thread.start();
		while (lua_thread.isRunning())
			QCoreApplication::processEvents();
		logInfo("MainWindow::showEvent : finished reading config");
	}

	render();

	connect(m_viewer, SIGNAL(viewerResized(const QSize&)),
		this, SLOT(mainViewerResizedAction(const QSize&)));
	connect(m_viewer, SIGNAL(viewerHidden()),
			this, SLOT(mainViewerHiddenAction()));
	connect(m_previewWidget, SIGNAL(previewResized(const QSize&)),
			this, SLOT(previewResizedAction(const QSize&)));
}


void MainWindow::newFile()
{
	setFlameXML(DEFAULT_FLAME_XML);
	setCurrentFile("");
}


void MainWindow::open()
{
	QString fileName
			= QFileDialog::getOpenFileName(this, "Open a flame", lastDir,
		  "flam3 xml (*.flam *.flam3 *.flame);;All files (*)");
	if (!fileName.isEmpty())
	{
		lastDir = QFileInfo(fileName).dir().canonicalPath();
		if (loadFile(fileName))
			render();
	}
}


bool MainWindow::save()
{
	if (curFile.isEmpty()) {
		return saveAs();
	} else {
		return saveFile(curFile);
	}
}


bool MainWindow::saveAs()
{
	QString fileName = curFile;
	if (fileName.isEmpty())
		fileName = QString(".flam3");

	fileName = QFileDialog::getSaveFileName(this, "Save a flame",
				QDir(lastDir).absoluteFilePath(fileName),
				"flam3 xml (*.flam *.flam3 *.flame);;All files (*)");
	if (fileName.isEmpty())
		return false;

	lastDir = QFileInfo(fileName).dir().canonicalPath();
	return saveFile(fileName);
}


bool MainWindow::quickSave()
{
	QStringList filter;
	filter << "q*.flam3";
	QDir quickDir(m_directoryViewWidget->currentPath());
	QFileInfoList files = quickDir.entryInfoList(filter, QDir::Files,
			QDir::Name | QDir::Reversed);

	int pad = 8;
	QFileInfo file(quickDir, QString("q%1.flam3").arg(1, pad, 10, QChar('0')));

	if (files.size() > 0)
	{
		QRegExp r("q(\\d+).flam3");
		foreach (QFileInfo i, files)
		{
			QString lastname = i.fileName();
			if (r.exactMatch(lastname))
			{
				int n = r.cap(1).toInt();
				if (++n > 99999999)
					pad += 4;
				lastname = QString("q%1.flam3").arg(n, pad, 10, QChar('0'));
				file = QFileInfo(quickDir, lastname);
				break;
			}
		}
	}

	QString filename = file.absoluteFilePath();
	QString basename = file.baseName();
	QFileInfo imgFile = QFileInfo(file.dir(), basename + ".png");
	logInfo(QString("MainWindow::quickSave : quicksaving to file %1")
			.arg(filename));
	saveFile(filename);
	return saveImage(imgFile.absoluteFilePath());
}

bool MainWindow::saveImage(const QString& filename, int idx)
{
	QString fileName(filename);
	if (fileName.isEmpty())
	{
		QString imageName = curFile;
		if (imageName.isEmpty())
			imageName = "untitled.png";
		else
		{
			int pos = imageName.lastIndexOf('.');
			imageName.replace(pos, imageName.size() - pos, ".png");
		}
		fileName = QFileDialog::getSaveFileName(this,
			"Save an image", QDir(lastDir).absoluteFilePath(imageName));
	}

	if (fileName.isEmpty() || !QFileInfo(QFileInfo(fileName).path()).isWritable())
	{
		logInfo(QString("MainWindow::saveImage : couldn't save to file %1")
				.arg(fileName));
		return false;
	}

	lastDir = QFileInfo(fileName).dir().canonicalPath();
	logInfo(QString("MainWindow::saveImage : rendering to file %1")
			.arg(fileName));
	if (idx > -1)
		m_file_request.setGenome(genomes.data() + idx);
	else
		m_file_request.setGenome(genomes.selectedGenome());
	m_file_request.setName(fileName);
	m_file_request.setType(RenderRequest::File);
	m_rthread->render(&m_file_request);
	return true;
}

void MainWindow::importAction()
{
	QString fileName
			= QFileDialog::getOpenFileName(this, "Import genomes from a file", lastDir,
		  "flam3 xml (*.flam *.flam3 *.flame);;All files (*)");
	if (!fileName.isEmpty())
	{
		lastDir = QFileInfo(fileName).dir().canonicalPath();
		if (importGenome(fileName))
			render();
	}
}

void MainWindow::exportAction()
{
	QString fileName(curFile);
	if (fileName.isEmpty())
		fileName = QString(".flam3");
	else
		fileName.replace(QRegExp("\\.flam[3e]?$"),
			QString("_%1.flam3").arg(genomes.selectedIndex()));

	fileName = QFileDialog::getSaveFileName(this, "Save a the current genome",
				QDir(lastDir).absoluteFilePath(fileName),
				"flam3 xml (*.flam *.flam3 *.flame);;All files (*)");
	if (!fileName.isEmpty())
	{
		lastDir = QFileInfo(fileName).dir().canonicalPath();
		exportGenome(fileName, genomes.selectedIndex());
	}
}

void MainWindow::about()
{
	QMessageBox b(QMessageBox::NoIcon, tr("About Qosmic"),
		QString("<table align=\"center\"><tr><td>"
				"<img src=\":/icons/qosmic.xpm\">&nbsp; </td>"
				"<td><b>Qosmic</b> (%1) </p>"
				"<p> A cosmic recursive flame fractal editor</p>"
				"<p>Written by: David Bitseff</p>"
				"<p>Thanks to:<li>- Scott Draves for his flam3 library"
				"<li>- Mark James for his <a href=\"http://www.famfamfam.com/lab/icons/silk/\">Silk icon set</a>"
				"<li>- The Apophysis developers for their interface design</p>"
				"<p>This program is licensed under the GNU GPL Version 2"
				"</p></td></tr></table>").arg(QOSMIC_VERSION),QMessageBox::Ok,this);
	b.exec();
}


void MainWindow::createActions()
{
	logInfo("MainWindow::createActions : creating actions");
	newAct = new QAction(QIcon(":icons/silk/application.xpm"), tr("Reset"), this);
	newAct->setShortcut(tr("Ctrl+N"));
	newAct->setStatusTip(tr("Load default genome"));
	connect(newAct, SIGNAL(triggered()), this, SLOT(newFile()));

	openAct = new QAction(QIcon(":icons/silk/folder.xpm"), tr("&Open..."), this);
	openAct->setShortcut(tr("Ctrl+O"));
	openAct->setStatusTip(tr("Open an existing file"));
	connect(openAct, SIGNAL(triggered()), this, SLOT(open()));

	saveAct = new QAction(QIcon(":icons/silk/disk.xpm"), tr("&Save"), this);
	saveAct->setShortcut(tr("Ctrl+S"));
	saveAct->setStatusTip(tr("Save the flame to disk"));
	connect(saveAct, SIGNAL(triggered()), this, SLOT(save()));

	saveAsAct = new QAction(QIcon(":icons/silk/page_save.xpm"),tr("Save &As..."), this);
	saveAsAct->setStatusTip(tr("Save the flame under a new name"));
	connect(saveAsAct, SIGNAL(triggered()), this, SLOT(saveAs()));

	saveImageAct = new QAction(QIcon(":icons/silk/picture_save.xpm"),tr("Render to &file..."), this);
	saveImageAct->setStatusTip(tr("Save an image of current flame"));
	connect(saveImageAct, SIGNAL(triggered()), this, SLOT(saveImage()));

	quickSaveAct = new QAction(QIcon(":icons/silk/disk_multiple.xpm"),tr("Q&uicksave flame..."), this);
	quickSaveAct->setStatusTip(tr("Quickly save file and image of current flame"));
	connect(quickSaveAct, SIGNAL(triggered()), this, SLOT(quickSave()));

	for (int i = 0; i < NumRecentFiles; ++i) {
		recentFileActions[i] = new QAction(this);
		recentFileActions[i]->setVisible(false);
		connect(recentFileActions[i], SIGNAL(triggered()),
				this, SLOT(openRecentFile()));
	}

	exitAct = new QAction(QIcon(":icons/silk/cross.xpm"),tr("&Quit"), this);
	exitAct->setShortcut(tr("Ctrl+Q"));
	exitAct->setStatusTip(tr("Quit the application"));
	connect(exitAct, SIGNAL(triggered()), this, SLOT(close()));

	killAct = new QAction(QIcon(":icons/silk/stop.xpm"),tr("&Stop rendering"), this);
	killAct->setShortcut(tr("Ctrl+K"));
	killAct->setStatusTip(tr("Stop rendering"));
	connect(killAct, SIGNAL(triggered()), m_rthread, SLOT(killAll()));

	randomAct = new QAction(QIcon(":icons/silk/star.xpm"),tr("&Random Flame"), this);
	randomAct->setShortcut(tr("Ctrl+R"));
	randomAct->setStatusTip(tr("Generate a random xform set"));
	connect(randomAct, SIGNAL(triggered()), this, SLOT(randomizeGenomeAction()));

	rescaleAct = new QAction(QIcon(":icons/silk/shape_group.xpm"), tr("Rescale"), this);
	rescaleAct->setShortcut(tr("Ctrl+A"));
	rescaleAct->setStatusTip(tr("Rescale the editor"));
	connect(rescaleAct, SIGNAL(triggered()), m_xfeditor, SLOT(autoScale()));

	QAction* a = new QAction(tr("Select next triangle"), this);
	a->setShortcut(tr("F2"));
	addAction(a);
	connect(a, SIGNAL(triggered()), m_xfeditor, SLOT(selectNextTriangle()));

	a = new QAction(tr("Select previous triangle"), this);
	a->setShortcut(tr("F1"));
	addAction(a);
	connect(a, SIGNAL(triggered()), m_xfeditor, SLOT(selectPreviousTriangle()));

	undoAct = new QAction(QIcon(":icons/silk/arrow_undo.xpm"), tr("Undo"), this);
	undoAct->setShortcut(tr("F4"));
	undoAct->setStatusTip(tr("Undo"));
	addAction(undoAct);
	connect(undoAct, SIGNAL(triggered()), this, SLOT(undo()));

	redoAct = new QAction(QIcon(":icons/silk/arrow_redo.xpm"), tr("Redo"), this);
	redoAct->setShortcut(tr("Shift+F4"));
	redoAct->setStatusTip(tr("Redo"));
	addAction(redoAct);
	connect(redoAct, SIGNAL(triggered()), this, SLOT(redo()));

	importAct = new QAction(QIcon(":icons/silk/page_white_get.xpm"),
		tr("Import genomes"), this);
	importAct->setStatusTip(tr("Import genomes"));
	addAction(importAct);
	connect(importAct, SIGNAL(triggered()), this, SLOT(importAction()));

	exportAct = new QAction(QIcon(":icons/silk/page_white_put.xpm"),
		tr("Export current genome"), this);
	exportAct->setStatusTip(tr("Export the current genome"));
	addAction(exportAct);
	connect(exportAct, SIGNAL(triggered()), this, SLOT(exportAction()));

	fileToolBar = addToolBar(tr("File"));
	fileToolBar->setObjectName("FileToolBar");
	showFileBarAct = new QAction(tr("Show &File Toolbar"), this);
	showFileBarAct->setStatusTip(tr("Show the file toolbar"));
	connect(showFileBarAct, SIGNAL(triggered()), fileToolBar, SLOT(show()));

	editToolBar = addToolBar(tr("Widgets"));
	editToolBar->setObjectName("WidgetsToolBar");
	showEditBarAct = new QAction(tr("Show &Widgets Toolbar"), this);
	showEditBarAct->setStatusTip(tr("Show the widgets toolbar"));
	connect(showEditBarAct, SIGNAL(triggered()), editToolBar, SLOT(show()));

	extraToolBar = addToolBar(tr("Edit"));
	extraToolBar->setObjectName("EditToolBar");
	showExtraBarAct = new QAction(tr("Show &Edit Toolbar"), this);
	showExtraBarAct->setStatusTip(tr("Show the edit toolbar"));
	connect(showExtraBarAct, SIGNAL(triggered()), extraToolBar, SLOT(show()));

	aboutAct = new QAction(QIcon(":icons/silk/information.xpm"),tr("&About"), this);
	aboutAct->setStatusTip(tr("Show the application's About box"));
	connect(aboutAct, SIGNAL(triggered()), this, SLOT(about()));

	aboutQtAct = new QAction(tr("About &Qt"), this);
	aboutQtAct->setStatusTip(tr("Show the Qt library's About box"));
	connect(aboutQtAct, SIGNAL(triggered()), qApp, SLOT(aboutQt()));

}


void MainWindow::createMenus()
{
	logInfo("MainWindow::createMenus : adding menus");
	fileMenu = menuBar()->addMenu(tr("&File"));
	fileMenu->addAction(newAct);
	fileMenu->addAction(openAct);
	fileMenu->addAction(saveAct);
	fileMenu->addAction(saveAsAct);
	fileMenu->addAction(saveImageAct);
	separatorAct = fileMenu->addSeparator();
	for (int i = 0; i < NumRecentFiles; ++i)
		fileMenu->addAction(recentFileActions[i]);
	updateRecentFileActions();
	fileMenu->addSeparator();
	fileMenu->addAction(exitAct);

	editMenu = menuBar()->addMenu(tr("&Edit"));
	editMenu->addAction(undoAct);
	editMenu->addAction(redoAct);
	editMenu->addAction(rescaleAct);
	editMenu->addAction(randomAct);
	editMenu->addAction(killAct);
	editMenu->addSeparator();
	editMenu->addAction(importAct);
	editMenu->addAction(exportAct);

	settingsMenu = menuBar()->addMenu(tr("&Widgets"));
	settingsMenu->addAction(showFileBarAct);
	settingsMenu->addAction(showEditBarAct);
	settingsMenu->addAction(showExtraBarAct);
	settingsMenu->addSeparator();

	helpMenu = menuBar()->addMenu(tr("&Info"));
	helpMenu->addAction(aboutAct);
	helpMenu->addAction(aboutQtAct);
}

void MainWindow::updateRecentFileActions()
{
	QSettings settings;
	QStringList files = settings.value("recentFilesList").toStringList();

	int numRecentFiles = qMin(files.size(), (int)NumRecentFiles);

	for (int i = 0; i < numRecentFiles; ++i) {
		QString text = tr("&%1 %2").arg(i + 1).arg(strippedName(files[i]));
		recentFileActions[i]->setText(text);
		recentFileActions[i]->setData(files[i]);
		recentFileActions[i]->setVisible(true);
	}
	for (int j = numRecentFiles; j < NumRecentFiles; ++j)
		recentFileActions[j]->setVisible(false);

	separatorAct->setVisible(numRecentFiles > 0);
}

void MainWindow::createToolBars()
{
	logInfo("MainWindow::createToolBars : creating toolbars");
	fileToolBar->addAction(newAct);
	fileToolBar->addAction(openAct);
	fileToolBar->addAction(saveAct);
	fileToolBar->addAction(saveAsAct);
	fileToolBar->addAction(saveImageAct);
	fileToolBar->addAction(quickSaveAct);

	QAction* action = qobject_cast<QDockWidget*>(m_viewer->parentWidget())->toggleViewAction();
	action->setIcon(QIcon(":icons/silk/monitor.xpm"));
	action->setStatusTip(tr("Viewer"));
	action->setShortcut(tr("F3"));
	editToolBar->addAction(action);

	action = qobject_cast<QDockWidget*>(m_cameraSettingsWidget->parentWidget())->toggleViewAction();
	action->setIcon(QIcon(":icons/silk/camera.xpm"));
	action->setStatusTip(tr("Camera"));
	editToolBar->addAction(action);

	action = qobject_cast<QDockWidget*>(m_selectTriangleWidget->parentWidget())->toggleViewAction();
	action->setIcon(QIcon(":icons/silk/shape_move_forwards.xpm"));
	action->setStatusTip(tr("Triangles"));
	editToolBar->addAction(action);

	action = qobject_cast<QDockWidget*>(m_imageSettingsWidget->parentWidget())->toggleViewAction();
	action->setIcon(QIcon(":icons/silk/image_edit.xpm"));
	action->setStatusTip(tr("Image Quality"));
	editToolBar->addAction(action);

	action = qobject_cast<QDockWidget*>(m_previewWidget->parentWidget())->toggleViewAction();
	action->setIcon(QIcon(":icons/silk/photo.xpm"));
	action->setStatusTip(tr("Previewer"));
	editToolBar->addAction(action);

	action = qobject_cast<QDockWidget*>(m_editTriangleWidget->parentWidget())->toggleViewAction();
	action->setIcon(QIcon(":icons/silk/shape_rotate_clockwise.xpm"));
	action->setStatusTip(tr("Triangle Editing"));
	editToolBar->addAction(action);

	action = qobject_cast<QDockWidget*>(m_variationsWidget->parentWidget())->toggleViewAction();
	action->setIcon(QIcon(":icons/silk/table.xpm"));
	action->setStatusTip(tr("Triangle Variations"));
	editToolBar->addAction(action);

	action = qobject_cast<QDockWidget*>(m_colorSettingsWidget->parentWidget())->toggleViewAction();
	action->setIcon(QIcon(":icons/silk/color_swatch.xpm"));
	action->setStatusTip(tr("Triangle Color"));
	editToolBar->addAction(action);

	action = qobject_cast<QDockWidget*>(m_colorBalanceWidget->parentWidget())->toggleViewAction();
	action->setIcon(QIcon(":icons/silk/contrast.xpm"));
	action->setStatusTip(tr("Color Settings"));
	editToolBar->addAction(action);

	action = qobject_cast<QDockWidget*>(m_paletteEditor->parentWidget())->toggleViewAction();
	action->setIcon(QIcon(":icons/silk/palette.xpm"));
	action->setStatusTip(tr("Palettes"));
	editToolBar->addAction(action);

	action = qobject_cast<QDockWidget*>(m_mutations->parentWidget())->toggleViewAction();
	action->setIcon(QIcon(":icons/silk/application_view_tile.xpm"));
	action->setStatusTip(tr("Mutations"));
	editToolBar->addAction(action);

	action = qobject_cast<QDockWidget*>(m_genomeSelectWidget->parentWidget())->toggleViewAction();
	action->setIcon(QIcon(":icons/silk/layers.xpm"));
	action->setStatusTip(tr("Genome List"));
	editToolBar->addAction(action);

	action = qobject_cast<QDockWidget*>(m_triangleDensityWidget->parentWidget())->toggleViewAction();
	action->setIcon(QIcon(":icons/silk/chart_bar.xpm"));
	action->setStatusTip(tr("Triangle Densities"));
	editToolBar->addAction(action);

	action = qobject_cast<QDockWidget*>(m_directoryViewWidget->parentWidget())->toggleViewAction();
	action->setIcon(QIcon(":icons/silk/folder_explore.xpm"));
	action->setStatusTip(tr("Directory Browser"));
	editToolBar->addAction(action);

	action = qobject_cast<QDockWidget*>(m_coordsWidget->parentWidget())->toggleViewAction();
	action->setIcon(QIcon(":icons/silk/text_columns.xpm"));
	action->setStatusTip(tr("Triangle Coordinates"));
	editToolBar->addAction(action);

	action = qobject_cast<QDockWidget*>(m_viewerPresetsWidget->parentWidget())->toggleViewAction();
	action->setIcon(QIcon(":icons/silk/images.xpm"));
	action->setStatusTip(tr("Viewer Presets"));
	editToolBar->addAction(action);

	action = qobject_cast<QDockWidget*>(m_scriptEditWidget->parentWidget())->toggleViewAction();
	action->setIcon(QIcon(":icons/silk/script.xpm"));
	action->setStatusTip(tr("Script Editor"));
	editToolBar->addAction(action);


	extraToolBar->addAction(undoAct);
	extraToolBar->addAction(redoAct);
	extraToolBar->addAction(rescaleAct);
	extraToolBar->addAction(randomAct);
	extraToolBar->addAction(killAct);
}


void MainWindow::createStatusBar()
{
	statusBar()->addWidget(&coordsLabel);
	statusBar()->addPermanentWidget(m_statusWidget);
	statusBar()->showMessage(tr("Ready"), 1000);
}


void MainWindow::readSettings()
{
	logInfo("MainWindow::readSettings : reading settings");

	QSettings settings;
	bool flag = settings.value("editor_grid_visible", false).toBool();
	m_xfeditor->setGridVisible(flag);
	QString color = settings.value("editor_grid_color", "#999999").toString();
	m_xfeditor->setGridColor(QColor(color));
	color = settings.value("editor_bg_color", "#000000").toString();
	m_xfeditor->setbgColor(QColor(color));

	lastDir = settings.value("palettes_last_directory", "").toString();
	m_paletteEditor->setLastBrowseDir(lastDir);
	lastDir = settings.value("last_directory", "").toString();
	m_directoryViewWidget->setCurrentPath(lastDir);

	QSize size = settings.value("size_preview_image", QSize(120, 100)).toSize();
	m_preview_request.setSize(size);

	QPoint pos = settings.value("pos", QPoint(200, 200)).toPoint();
	size = settings.value("size", QSize(580, 500)).toSize();
	resize(size);
	move(pos);

	pos  = settings.value("pos_viewer", QPoint(10, 200)).toPoint();
	size = settings.value("size_viewer", QSize(200, 180)).toSize();
	m_viewer->parentWidget()->resize(size);
	m_viewer->parentWidget()->move(pos);

	QVariant v = settings.value("editor_basis");
	if (!v.isNull() && v.convert(QVariant::Matrix))
		m_xfeditor->setCoordTransform(v.value<QMatrix>());
}


void MainWindow::writeSettings()
{
	logInfo("MainWindow::writeSettings : writing settings");
	QSettings settings;
	settings.setValue("state_mainwindow", saveState());

	settings.setValue("pos", pos());
	settings.setValue("size", size());

	settings.setValue("size_preview_image", m_previewWidget->getPreviewSize());

	settings.setValue("pos_viewer", m_viewer->parentWidget()->pos());
	settings.setValue("size_viewer", m_viewer->parentWidget()->size());
	settings.setValue("visible_viewer", m_viewer->parentWidget()->isVisible());

	settings.setValue("last_directory", m_directoryViewWidget->currentPath());
	settings.setValue("palettes_last_directory", m_paletteEditor->lastBrowseDir());

	settings.setValue("editor_grid_visible", m_xfeditor->gridVisible());
	settings.setValue("editor_grid_color", m_xfeditor->gridColor().name());
	settings.setValue("editor_bg_color", m_xfeditor->bgColor().name());
	settings.setValue("editor_basis", QVariant::fromValue(m_xfeditor->transform()));
}


void MainWindow::loadGenomeList(flam3_genome* gen, int ncps)
{
	if (genomes.size() > 0)
	{
		// free the heap mem used by xforms in existing genomes
		int n = 0;
		foreach (flam3_genome g, genomes)
		{
			if (g.num_xforms > 0)
			{
				logFine(QString("MainWindow::loadGenomeList : reclaiming %1 ")
						.append("xforms from genome %2")
					   .arg(g.num_xforms).arg(n));
				delete g.xform;
			}
			n++;
		}
	}
	genomes.clear();
	undo_ring.clear();
	appendFlam3ToGenome(gen, ncps);
	f.ngenomes = ncps;
	f.genomes = genomes.data();
	m_genomeSelectWidget->setSelectedGenome(0); // sets the selectedGenome
	undo_ring.append(f.genomes);
}

void MainWindow::setFlameXML(const QString& s)
{
	int ncps = 0;
	flam3_genome* in;
	 if (s.isEmpty())
		 in = Util::read_xml_string(DEFAULT_FLAME_XML, &ncps);
	 else
		 in = Util::read_xml_string(s, &ncps);
	logInfo(QString("MainWindow::setFlameXML : found %1 genome").arg(ncps));
	loadGenomeList(in, ncps);
	reset();
	m_previewWidget->parentWidget()->show();
	render();
}

void MainWindow::openRecentFile()
{
	QAction* action = qobject_cast<QAction*>(sender());
	if (action)
	{
		loadFile(action->data().toString());
		render();
	}
}


bool MainWindow::loadFile(const QString& fname)
{
	logInfo(QString("MainWindow::loadFile : opening %1").arg(fname));
	int ncps = 0;
	flam3_genome* in;
	if (readFlam3File(fname, &in, &ncps))
	{
		logInfo(QString("MainWindow::loadFile : found %1 genomes").arg(ncps));
		loadGenomeList(in, ncps);
		setCurrentFile(fname);
		reset();
		statusBar()->showMessage(tr("File loaded"), 2000);
		emit mainWindowChanged();
		return true;
	}
	return false;
}


bool MainWindow::saveFile(const QString &fname)
{
	QFile file(fname);
	char* attrs = "";

	logInfo(QString("MainWindow::saveFile : saving '%1'").arg(fname));
	if (!file.open(QIODevice::WriteOnly))
	{
		if (m_dialogsEnabled)
			QMessageBox::warning(this, tr("Application error"),
									tr("Cannot write file %1\n").arg(fname));
		return false;
	}
	FILE* fd = fdopen(file.handle(), "w"); // check for error again?
	if (genomes.size() > 1)
	{
		fprintf(fd, "<qstack>\n");
		foreach (flam3_genome g, genomes)
		{
			int n = g.symmetry;
			g.symmetry = 0;
			flam3_print(fd, &g, attrs, 0);
			g.symmetry = n;
		}
		fprintf(fd, "</qstack>\n");
	}
	else
	{
		// always clear the genome symmetry since the parser modifies the
		// genome when reading this tag.
		int n = genomes[0].symmetry;
		genomes[0].symmetry = 0;
		flam3_print(fd, genomes.data(), attrs, 0);
		genomes[0].symmetry = n;
	}

	fclose(fd);
	file.close();
	setCurrentFile(fname);
	statusBar()->showMessage(tr("File saved"), 2000);
	return true;
}

bool MainWindow::importGenome(const QString& fname)
{
	logInfo(QString("MainWindow::importGenome : opening %1").arg(fname));
	int ncps = 0;
	flam3_genome* in;
	if (readFlam3File(fname, &in, &ncps))
	{
		logInfo(QString("MainWindow::importGenome : found %1 genomes")
			.arg(ncps));
		appendFlam3ToGenome(in, ncps);
		reset();
		m_genomeSelectWidget->reset();
		emit mainWindowChanged();
		return true;
	}
	return false;
}

void MainWindow::appendFlam3ToGenome(flam3_genome* gen, int ncps)
{
	// copy the genomes into the currint list
	for (int n = 0 ; n < ncps ; n++, gen++)
	{
		gen->symmetry = 1; // clear genome symmetry flag
		// copy out the xforms too.  the flam3_parse_ function will
		// lose/overwrite these if it's called again.
		flam3_xform* xforms = new flam3_xform[gen->num_xforms];
		memset(xforms, 0, gen->num_xforms*sizeof(flam3_xform));
		memcpy(xforms, gen->xform, gen->num_xforms*sizeof(flam3_xform));
		// now free the xforms, they aren't recycled (lost) by libflam3.a
		free(gen->xform);
		gen->xform = xforms;
		genomes.append(*gen);
	}
}

bool MainWindow::exportGenome(const QString& fname, int index)
{
	QFile file(fname);
	char* attrs = "";
	logInfo(QString("MainWindow::exportGenome : saving to '%1'").arg(fname));
	if (!file.open(QIODevice::WriteOnly))
	{
		if (m_dialogsEnabled)
			QMessageBox::warning(this, tr("Application error"),
				tr("Cannot write file to %1\n").arg(fname));
		return false;
	}
	FILE* fd = fdopen(file.handle(), "w"); // check for error again?
	// always clear the genome symmetry since the parser modifies the
	// genome when reading this tag.
	int n = genomes[index].symmetry;
	genomes[index].symmetry = 0;
	flam3_print(fd, genomes.data() + index, attrs, 0);
	genomes[index].symmetry = n;
	fclose(fd);
	file.close();
	statusBar()->showMessage(tr("File saved"), 2000);
	return true;
}

bool MainWindow::readFlam3File(const QString& fname, flam3_genome** in, int* ncps)
{
	QFile file(fname);
	if (!file.open(QIODevice::ReadOnly))
	{
		if (m_dialogsEnabled)
			QMessageBox::warning(this, tr("Application error"),
				tr("Cannot open file %1\n").arg(fname));
		return false;
	}
	FILE* fd = fdopen(file.handle(), "r");
	// 'in' points to a static defined in flam3.c
	*in = flam3_parse_from_file(fd, fname.toAscii().data(), 1, ncps);
	fclose(fd);
	file.close();
	if (*ncps < 1)
	{
		if (m_dialogsEnabled)
			QMessageBox::warning(this, tr("Application error"),
				tr("Couldn't parse file %1\n").arg(fname));
		return false;
	}
	return true;
}

void MainWindow::setCurrentFile(const QString &fileName)
{
	curFile = fileName;

	QString shownName;
	if (curFile.isEmpty())
		shownName = "untitled.flam3";
	else
	{
		shownName = strippedName(curFile);
		if (shownName != ".qosmic.flam3")
		{
			QSettings settings;
			QStringList files = settings.value("recentFilesList").toStringList();
			QString filePath(QFileInfo(fileName).absoluteFilePath());
			files.removeAll(filePath);
			files.prepend(filePath);
			while (files.size() > NumRecentFiles)
				files.removeLast();
			settings.setValue("recentFilesList", files);
			updateRecentFileActions();
		}
	}
	setWindowTitle(tr("%1 - %2[*]").arg(tr("qosmic")).arg(shownName));
}


QString MainWindow::strippedName(const QString &fullFileName)
{
	return QFileInfo(fullFileName).fileName();
}


MainWindow::~MainWindow()
{
	// and so on...
	if (m_fileViewer)
		delete m_fileViewer;
	delete m_directoryViewWidget;
	delete m_mutations;
	delete m_paletteEditor;
	delete m_viewer;
	delete m_xfeditor;
	delete m_rthread;
}


// called when the button is held
void MainWindow::colorSelected(double idx)
{
	m_xfeditor->colorChangedAction(idx);
	render();
}


void MainWindow::paletteChangedAction()
{
	// copy palette to g.palette
	m_paletteEditor->getPalette(genomes.selectedGenome()->palette);
	m_colorSettingsWidget->reset();
	m_colorBalanceWidget->reset();
	m_xfeditor->colorChangedAction(selectedTriangle->xform()->color[0]);
	render();
}

void MainWindow::paletteHueChangedAction()
{
	m_paletteEditor->setPalette(genomes.selectedGenome()->palette);
	m_colorSettingsWidget->reset();
	m_xfeditor->colorChangedAction(selectedTriangle->xform()->color[0]);
	render();
}

void MainWindow::mainViewerResizedAction(const QSize& s)
{
	if (s.isValid())
	{
		logFine(QString("MainWindow::mainViewerResizedAction : new size %1,%2")
				.arg(s.width()).arg(s.height()));
		renderViewer();
	}
}

void MainWindow::previewResizedAction(const QSize& s)
{
	logFine(QString("MainWindow::previewResizedAction : new size %1,%2")
			.arg(s.width()).arg(s.height()));
	renderPreview();
}


void MainWindow::render()
{
	renderPreview();
	renderViewer();
	genome_modified_flag = true;
}

void MainWindow::renderViewer()
{
	if (m_viewer->isVisible())
	{
		flam3_genome* render_genome;
		if (m_triangleDensityWidget->hasMergedGenome())
			render_genome = m_triangleDensityWidget->getMergedGenome();
		else
			render_genome = genomes.selectedGenome();

		m_viewer_request.setGenome(render_genome);
		m_viewer_request.setSize(m_viewer->getViewerSize());
		m_rthread->render(&m_viewer_request);
	}
}

void MainWindow::renderPreview(int idx)
{
	if (m_previewWidget->isVisible())
	{
		flam3_genome* render_genome;
		if (idx > -1)
			render_genome = genomes.data() + idx;
		else if (m_triangleDensityWidget->hasMergedGenome())
			render_genome = m_triangleDensityWidget->getMergedGenome();
		else
			render_genome = genomes.selectedGenome();

		m_preview_request.setGenome(render_genome);
		m_preview_request.setSize(m_previewWidget->getPreviewSize());
		m_rthread->render(&m_preview_request);
	}
}


void MainWindow::randomizeGenomeAction()
{
	int ivar = flam3_variation_random;
	flam3_genome s;
	flam3_genome* g = genomes.selectedGenome();
	s = *g; // no need to copy xforms
	flam3_random(g, &ivar, 1, 0, 0);
	// restore the camera/image settings which flam3_random resets
	g->width                  = s.width;
	g->height                 = s.height;
	g->estimator              = s.estimator;
	g->estimator_curve        = s.estimator_curve;
	g->estimator_minimum      = s.estimator_minimum;
	g->sample_density         = s.sample_density;
	g->pixels_per_unit        = s.pixels_per_unit;
	g->spatial_filter_radius  = s.spatial_filter_radius;
	g->nbatches               = s.nbatches;
	g->symmetry               = s.symmetry;
	g->ntemporal_samples      = s.ntemporal_samples;
	m_paletteEditor->setPalette(g->palette);
	m_xfeditor->reset();
	render();
}


void MainWindow::flam3FileSelectedAction(const QString& name, bool append)
{
	if (append)
		importGenome(name);
	else
		loadFile(name);
	render();
}

void MainWindow::mutationSelectedSlot(flam3_genome* newg)
{
	flam3_genome* g = genomes.selectedGenome();
	if (g->num_xforms != newg->num_xforms)
	{
		g->num_xforms = newg->num_xforms;
		g->xform = (flam3_xform*)realloc(g->xform,
					sizeof(flam3_xform) * newg->num_xforms);
	}
	memcpy(g->xform, newg->xform, g->num_xforms * sizeof(flam3_xform));
	m_xfeditor->reset();
	render();
}

void MainWindow::reset()
{
	m_viewer_request.setGenome(genomes.selectedGenome());
	m_preview_request.setGenome(genomes.selectedGenome());

	logFine("MainWindow::reset : resetting figure editor");
	// the colorsettings and variations are updated when a triangle is selected
	m_xfeditor->reset();
	logFine("MainWindow::reset : resetting palettes");
	m_paletteEditor->setPalette(genomes.selectedGenome()->palette);
	logFine("MainWindow::reset : resetting color balance");
	m_colorBalanceWidget->reset();
	logFine("MainWindow::reset : resetting image settings");
	m_imageSettingsWidget->reset();
	logFine("MainWindow::reset : resetting camera settings");
	m_cameraSettingsWidget->reset();
}

void MainWindow::scriptFinishedSlot()
{
	reset();
	// reset the last selected genome
	m_genomeSelectWidget->setSelectedGenome(genomes.selectedIndex());
}



GenomeVector* MainWindow::genomeVector()
{
	return &genomes;
}

FigureEditor* MainWindow::xformEditor() const
{
	return m_xfeditor;
}


bool MainWindow::dialogsEnabled() const
{
	return m_dialogsEnabled;
}

void MainWindow::setDialogsEnabled(bool value)
{
	m_dialogsEnabled = value;
}

void MainWindow::selectGenomeSlot(int idx)
{
	logFine(QString("MainWindow::selectGenomeSlot : %1").arg(idx));
	reset();
	render();
}

void MainWindow::presetSelectedSlot()
{
	logFine("MainWindow::presetSelectedSlot : applying preset");
	m_imageSettingsWidget->reset();
	m_cameraSettingsWidget->reset();
	m_colorBalanceWidget->reset();
	render();
}

RenderThread* MainWindow::renderThread() const
{
	return m_rthread;
}

void MainWindow::setImageFormat(int format)
{
	m_rthread->setFormat(static_cast<RenderThread::ImageFormat>(format));
	render();
}

void MainWindow::mainViewerHiddenAction()
{
	// stop rendering the mainviewer if it's waiting for an image.
	if (m_rthread->isRendering() && m_rthread->current() == &m_viewer_request)
		m_rthread->stopRendering();
}

bool MainWindow::eventFilter(QObject* obj, QEvent* event)
{
	if (event->type() == QEvent::KeyPress
		&& static_cast<QKeyEvent*>(event)->key() == Qt::Key_Space)
	{
		undo_ring.append(genomes.selectedGenome());
		genome_modified_flag = false;
		return true;
	}
	return false;
}

void MainWindow::undo()
{
	if (genome_modified_flag)
		undo_ring.append(genomes.selectedGenome());

	if (undo_ring.atTail())
		statusBar()->showMessage(tr("last undo"), 1000);

	else
	{
		flam3_genome* old = undo_ring.prev();
		flam3_genome* g = genomes.selectedGenome();
		flam3_copy(g, old);
		reset();
		render();
	}
	genome_modified_flag = false;
}

void MainWindow::redo()
{
	if (undo_ring.atHead())
		statusBar()->showMessage(tr("last redo"), 1000);

	else
	{
		flam3_genome* old = undo_ring.next();
		flam3_genome* g = genomes.selectedGenome();
		flam3_copy(g, old);
		reset();
		render();
	}
	genome_modified_flag = false;
}

