/***************************************************************************
 *   Copyright (C) 2007 by David Bitseff                                   *
 *   dbitsef@zipcon.net                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <QDir>
#include <QPainter>
#include <QMessageBox>
#include <QFileDialog>

#include "paletteeditor.h"
#include "flam3util.h"
#include "logger.h"


PaletteEditor::PaletteEditor ( QWidget* parent ) : QWidget(parent)
{
	setupUi ( this );
	loadPalette(0);

	m_palettesView->setModel(&m_flamPalettes);
	m_browseView->setModel(&m_browsePalettes);
	hasUGR = false;

	connect(m_palettesView, SIGNAL(pressed(const QModelIndex&)),
			this, SLOT(paletteIndexChangedAction(const QModelIndex&)));
	connect(m_browseView, SIGNAL(pressed(const QModelIndex&)),
			this, SLOT(selectGradientAction(const QModelIndex&)));
	connect(m_rotateSlider, SIGNAL(valueChanged (int)),
			this, SLOT(paletteRotatedAction(int)));
	connect(m_openButton, SIGNAL(clicked(bool)),
			this, SLOT(openGradientAction(bool)));
}

void PaletteEditor::showEvent(QShowEvent* event)
{
	setPaletteView();
	buildPaletteSelector();
}


void PaletteEditor::loadPalette(int palette_idx)
{
	flam3_get_palette(palette_idx, p, 0.0);
}

void PaletteEditor::setPaletteView()
{
	QSize s = m_paletteLabel->maximumSize();
	QImage palette(s.width(), s.height(), QImage::Format_RGB32);
	QPainter painter(&palette);
	int rvalue = m_rotateSlider->value() ;
	int n = 0;
	for (int i = (255-rvalue) % 255 ; n < 256 ; i = (i + 1) % 255, n++)
	{
		painter.setPen(QColor::fromRgbF(p[i][0], p[i][1], p[i][2]));
		painter.drawLine(n, 0, n, s.height());
	}
	m_paletteLabel->setPixmap(QPixmap::fromImage(palette));
}


void PaletteEditor::buildPaletteSelector()
{
	// only do this once, but it takes a while, so only when asked
	static bool built = false;
	if (built) return;
	built = true;
	logInfo("PaletteEditor::buildPaletteSelector : generating palettes");
	QSize s = m_palettesView->iconSize();
	for (int n = 0 ; n < npalettes ; n++)
	{
		flam3_palette p;
		flam3_get_palette(n, p, 0.0);
		QImage palette(s.width(), s.height(), QImage::Format_RGB32);
		QPainter painter(&palette);
		for (int i = 0 ; i < 256 ; i++)
		{
			painter.setPen(QColor::fromRgbF(p[i][0], p[i][1], p[i][2]));
			painter.drawLine(i, 0, i, s.height());
		}
		m_flamPalettes.addGradient(QPixmap::fromImage(palette));
	}
	if (!m_lastBrowseDir.isEmpty())
		openGradientAction(true);

}


void PaletteEditor::paletteIndexChangedAction(const QModelIndex& idx)
{
	if (!idx.isValid()) return;
	loadPalette(idx.row());
	setPaletteView();
	emit paletteChanged();
}

void PaletteEditor::paletteRotatedAction(int idx)
{
	setPaletteView();
	emit paletteChanged();
}

// i'd hoped to avoid this type of call
void PaletteEditor::getPalette(flam3_palette in)
{
	int rvalue = m_rotateSlider->value() ;
	int n = 0;
	for (int i = (255-rvalue) % 255 ; n < 256 ; i = (i + 1) % 255, n++)
	{
		in[n][0] = p[i][0];
		in[n][1] = p[i][1];
		in[n][2] = p[i][2];
	}
}

// and it's evil twin!
void PaletteEditor::setPalette(flam3_palette in)
{
	for (int n = 0 ; n < 256 ; n++)
	{
		p[n][0] = in[n][0];
		p[n][1] = in[n][1];
		p[n][2] = in[n][2];
	}
	setPaletteView();
	m_rotateSlider->blockSignals(true);
	m_rotateSlider->setSliderPosition(0);
	m_rotateSlider->blockSignals(false);
}

// select a gradient from the browse list
void PaletteEditor::selectGradientAction(const QModelIndex& idx)
{
	if (idx.isValid())
	{
		if (hasUGR)
		{
			logFine(QString("PaletteEditor::selectGradientAction : UGR %1")
								.arg(idx.row()));
			setPalette(ugrList[idx.row()].pa);
			emit paletteChanged();
		}
		else
		{
			QFileInfo file = m_browseFileList.at(idx.row());
			logFine(QString("PaletteEditor::selectGradientAction : selecting %1")
								.arg(idx.row()));
			if ((file.suffix() == "ggr" && loadGIMPGradient(file, p))
			||  (file.suffix() == "cpt" && loadCPTGradient(file, p)))
			{
				setPaletteView();
				emit paletteChanged();
			}
			else
				QMessageBox::warning(this, tr("Qosmic"),
						QString("The file %1 couldn't be opened.")
								.arg(file.absoluteFilePath()),
									 QMessageBox::NoButton);
		}
	}
}

// open a and parse a directory containing gimp gradient files
void PaletteEditor::openGradientAction(bool t)
{
	logFine(QString("PaletteEditor::openGradientAction : %1").arg(t));
	QString openDir;
	if (t)
	{
		openDir = m_lastBrowseDir;
		m_lastBrowseDir.clear();
	}
	else
	{
		openDir = QFileDialog::getOpenFileName(this,
			"Select a gradient file to load",
			QFileInfo(m_lastBrowseDir).absoluteFilePath(),
			tr("Gradient Files (*.cpt *.ggr *.ugr)"));
	}

	if (openDir.isEmpty())
		return;

	if (openDir.indexOf(QRegExp(".*\\.ugr$")) >= 0)
	{
		logFine(QString("PaletteEditor::openGradientAction : ugr %1").arg(openDir));
		QFileInfo ugrFile(openDir);
		if (ugrFile.exists())
		{
			loadUGRGradients(ugrFile);
			QSize s = m_browseView->iconSize();
			QImage palette(s.width(), s.height(), QImage::Format_RGB32);
			QPainter painter(&palette);
			m_browsePalettes.clear();
			foreach (flam3_palette_t t, ugrList)
			{
				for (int n = 0 ; n < 256 ; n++)
				{
					painter.setPen(QColor::fromRgbF(
								t.pa[n][0], t.pa[n][1], t.pa[n][2]));
					painter.drawLine(n, 0, n, s.height());
				}
				m_browsePalettes.addGradient(QPixmap::fromImage(palette));
			}
			hasUGR=true;
			m_lastBrowseDir = openDir;
			m_browseLineEdit->setText(m_lastBrowseDir);
		}
	}
	else
	{
		logFine(QString("PaletteEditor::openGradientAction : cpt,ggr %1")
				.arg(openDir));
		QFileInfo data(openDir);
		m_lastBrowseDir = openDir;
		if (!data.isDir())
			data = QFileInfo(data.absolutePath());

		QDir dir(data.absoluteFilePath());
		QStringList filters;
		filters << "*.ggr" << "*.cpt";
		m_browseFileList = dir.entryInfoList(filters, QDir::Files, QDir::Name);
		m_browsePalettes.clear();
		m_browseLineEdit->setText(data.absoluteFilePath());
		foreach (QFileInfo i, m_browseFileList)
		{
			flam3_palette pa;
			if ((i.suffix() == "ggr" && loadGIMPGradient(i, pa))
			||  (i.suffix() == "cpt" && loadCPTGradient(i, pa)))
			{
				QSize s = m_browseView->iconSize();
				QImage palette(s.width(), s.height(), QImage::Format_RGB32);
				QPainter painter(&palette);
				for (int n = 0 ; n < 256 ; n++)
				{
					painter.setPen(QColor::fromRgbF(pa[n][0], pa[n][1], pa[n][2]));
					painter.drawLine(n, 0, n, s.height());
				}
				m_browsePalettes.addGradient(QPixmap::fromImage(palette));
				logFinest(QString("PaletteEditor::openGradientAction : adding %1")
					.arg(i.absoluteFilePath()));
			}
			else
			{
				logWarn(QString("PaletteEditor::openGradientAction : skipping %1")
					.arg(i.absoluteFilePath()));
				m_browseFileList.removeAll(i);
			}
		}
		ugrList.clear();
		hasUGR=false;
	}
}


bool PaletteEditor::loadUGRGradients(QFileInfo& file)
{
	logInfo(QString("PaletteEditor::loadUGRGradients : opening %1")
			.arg(file.fileName()));
	QFile data(file.absoluteFilePath());
	if (!data.open(QFile::ReadOnly))
		return false;
	QTextStream is(&data);
	
	const double part = 256.0; // palette entries
	const double div  = part / 400; // entries / range
	QRegExp rx("index=(\\d+)\\s+color=(\\d+)");
	ugrList.clear();

	while (!is.atEnd())
	{
		flam3_palette pa;
		QString buf = is.readLine();
		// look for the first index
		while (! (rx.indexIn(buf) >= 0 || is.atEnd()))
			buf = is.readLine();

		int pidx = 0;
		while (( rx.indexIn(buf) >= 0 ) && !is.atEnd())
		{
			int idx   = rx.cap(1).toInt();
			int color = rx.cap(2).toInt();
			int j = (int)((double)idx*div);
			int b = (color >> 16) & 0xff;
			int g = (color >> 8) & 0xff;
			int r = color & 0xff;
			while (pidx <= j)
			{
				pa[pidx][0] = (double)r/255.;
				pa[pidx][1] = (double)g/255.;
				pa[pidx][2] = (double)b/255.;
				pidx++;
			}

			logFinest(QString("PaletteEditor::openUGRGradients : j=%1: %2,%3,%4")
				.arg(j).arg(pa[j][0]).arg(pa[j][1]).arg(pa[j][2]));
			
			buf = is.readLine();	
		}
		ugrList.append(flam3_palette_t(pa));

	}

	return true;
}

bool PaletteEditor::loadGIMPGradient(QFileInfo& file, flam3_palette pa)
{

	logFinest(QString("PaletteEditor::loadGIMPGradient : %1")
			.arg(file.fileName()));
	QFile data(file.absoluteFilePath());
	if (!data.open(QFile::ReadOnly))
		return false;
	QTextStream is(&data);

	int cpoints;
	QString buf = is.readLine();
	if (buf != "GIMP Gradient")
		return false;

	is >> buf;
	if (buf == "Name:")
		buf = is.readLine(); // buf now contains name of gradient
	is >> cpoints;
	for (int n = 0 ; n < cpoints ; n++)
	{
		double leftc,midc,rightc,leftr,leftg,leftb,lefta;
		double rightr,rightg,rightb,righta;
		int buckets;
		is >> leftc >> midc >> rightc
				>> leftr >> leftg >> leftb >> lefta
				>> rightr >> rightg >> rightb >> righta
				>> buckets >> buckets;
		int pstart = (int)(leftc * 256.);
		int pmid   = (int)(midc  * 256.);
		int pend   = (int)(rightc * 256.);
		buckets = pmid - pstart;
		double rdx = (rightr - leftr)/2.; double rinc = rdx / buckets;
		double gdx = (rightg - leftg)/2.; double ginc = gdx / buckets;
		double bdx = (rightb - leftb)/2.; double binc = bdx / buckets;
		for (int j = pstart ; j < pmid ; j++)
		{
			pa[j][0] = leftr;
			pa[j][1] = leftg;
			pa[j][2] = leftb;
			leftr += rinc;
			leftg += ginc;
			leftb += binc;
		}
		buckets = pend - pmid;
		rinc = rdx / buckets;
		ginc = gdx / buckets;
		binc = bdx / buckets;
		for (int j = pmid ; j < pend ; j++)
		{
			pa[j][0] = leftr;
			pa[j][1] = leftg;
			pa[j][2] = leftb;
			leftr += rinc;
			leftg += ginc;
			leftb += binc;
		}
	}
	return true;
}

// a silly cpt gradient file parser n' stuffer.  this thing
// is not very robust.  cpt files with problems/typos will break it!
bool PaletteEditor::loadCPTGradient(QFileInfo& file, flam3_palette pa)
{
	logFinest(QString("PaletteEditor::loadCPTGradient : parsing %1")
			.arg(file.fileName()));
	QFile data(file.absoluteFilePath());
	if (!data.open(QFile::ReadOnly))
		return false;
	QTextStream is(&data);

	QString buf = is.readLine();
	qint64 pos = is.pos();
	// skip comments
	while (buf.startsWith("#"))
	{
		pos = is.pos();
		buf = is.readLine();
	}
	is.seek(pos);

	double z0;
	double z1 = -1.0;
	double z0r, z0g, z0b, z1r, z1g, z1b;
	double offset = 0.;

	// first figure out the gradient range, and map
	// that into the palette's range (256)
	qint64 start_pos = pos;
	is >> z0 >> z0r >> z0g >> z0b >> z1 >> z1r >> z1g >> z1b;
	double idx_start = z0; // start position is always first
	if (idx_start != 0.0)
	{
		// make all positions >= 0
		offset = idx_start * -1.0;
		idx_start = 0.0;
	}
	QRegExp rx("^\\s*[BFM]\\s+\\d+");
	while (!is.atEnd())
	{
		pos = is.pos();
		buf = is.readLine();

		if (buf.isEmpty())
			continue;
		else if ( rx.indexIn(buf) != -1 )
			break;
		else
		{
			is.seek(pos);
			is >> z0 >> z0r >> z0g >> z0b >> z1 >> z1r >> z1g >> z1b;
			is.readLine();
		}
	}
	// end position is in the last entry
	double idx_end = z1 + offset;
	logFinest(QString("PaletteEditor::loadCPTGradient : idx range %1 %2")
			.arg(idx_start).arg(idx_end));
	// now go back and try again
	is.seek(start_pos);
	z1 = idx_start;

	const double part = 256.0; // palette entries
	const double div  = part / (idx_end - idx_start); // entries / range

	while (z1 < (idx_end - offset))
	{
		is >> z0 >> z0r >> z0g >> z0b >> z1 >> z1r >> z1g >> z1b;
		is.readLine();
		// move all indexes up, and use reals for rgb
		z0 += offset;  z1 += offset;
		z0r /= part; z0g /= part ; z0b /= part ;
		z1r /= part; z1g /= part ; z1b /= part ;

		int pstart = qRound(z0 * div);
		int pend   = qRound(z1 * div);
		double pdx = pend - pstart ;
		double rdx = (z1r - z0r) / pdx ;
		double gdx = (z1g - z0g) / pdx;
		double bdx = (z1b - z0b) / pdx;
		logFinest(QString("PaletteEditor::loadCPTGradient : z0 %1 \t %2 %3 %4")
			.arg(z0).arg(z0r).arg(z0g).arg(z0b));
		logFinest(QString("PaletteEditor::loadCPTGradient : z1 %1 \t %2 %3 %4")
			.arg(z1).arg(z0r).arg(z1g).arg(z1b));

		int j = pstart;
		for ( ; j < pend && j <= 255; j++)
		{
			pa[j][0] = qBound(0., z0r, 1.) ;
			pa[j][1] = qBound(0., z0g, 1.) ;
			pa[j][2] = qBound(0., z0b, 1.) ;
			logFinest(QString("PaletteEditor::loadCPTGradient : p[%1] \t %2 %3 %4")
				.arg(j).arg(p[j][0]).arg(p[j][1]).arg(p[j][2]));
			z0r += rdx ;
			z0g += gdx;
			z0b += bdx;
		}
		if (j >= 255) break;
	}
	return true;
}

QString PaletteEditor::lastBrowseDir() const
{
	return m_lastBrowseDir;
}

void PaletteEditor::setLastBrowseDir (const QString& value)
{
	m_lastBrowseDir = value;
}
