/***************************************************************************
 *   Copyright (C) 2007 by David Bitseff                                   *
 *   dbitsef@zipcon.net                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "variationswidget.h"
#include "varstablewidget.h"
#include "flam3util.h"
#include "logger.h"

#include <QTableWidget>
#include <QHeaderView>
#include <QStringList>
#include <QScrollBar>
#include <QFont>


VariationsWidget::VariationsWidget(GenomeVector* gen, QWidget* parent)
	: QWidget(parent), genome(gen)
{
	setupUi(this);

	// variations table
	m_variationsTable->horizontalHeader()->resizeSection(0,50);
	m_variationsTable->horizontalHeader()->setResizeMode(0,QHeaderView::Stretch);
	m_variationsTable->verticalHeader()->setDefaultSectionSize(20);
	QStringList headers;
	headers << QString("");
	QFont f = m_variationsTable->horizontalHeader()->font();
	f.setPointSize(8);
	m_variationsTable->horizontalHeader()->setFont(f);
	m_variationsTable->setHorizontalHeaderLabels(headers);

	QStringList vars(Util::flam3_variations().keys());
	m_variationsTable->setRowCount(vars.size());
	m_variationsTable->setVerticalHeaderLabels(vars);

	// color the labels for acceptable sheepserver variations
	for (int n = 0 ; n < vars.size() ; n++)
	{
		QTableWidgetItem* item = m_variationsTable->verticalHeaderItem(n);
		int varn = Util::variation_number(item->text());
		if (varn <= 22)
			item->setBackground(QColor(200,200,240));
	}

	// variables table
	m_variablesTable->horizontalHeader()->resizeSection(0,50);
	m_variablesTable->horizontalHeader()->setResizeMode(0,QHeaderView::Stretch);
	m_variablesTable->verticalHeader()->setDefaultSectionSize(20);
	f = m_variablesTable->horizontalHeader()->font();
	f.setPointSize(8);
	m_variablesTable->horizontalHeader()->setFont(f);
	m_variablesTable->setHorizontalHeaderLabels(headers);
	vars = Util::get_variable_names();
	m_variablesTable->setRowCount(vars.size());
	m_variablesTable->setVerticalHeaderLabels(vars);;

	connect(m_variationsTable, SIGNAL(cellChanged(int, int)),
			this, SLOT(variationEditedSlot(int,int)));
	connect(m_variablesTable, SIGNAL(cellChanged(int, int)),
			this, SLOT(variableEditedSlot(int,int)));
}

void VariationsWidget::setGenome(int g)
{
	genome_offset = g;
	updateFormData();
}

#define genome_ptr (genome->selectedGenome())

void VariationsWidget::updateFormData()
{
	logFiner("VariationsWidget::updateFormData : setting variations");
	static bool no_items = true;
	m_variationsTable->blockSignals(true);
	QStringList vars(Util::flam3_variations().keys());
	if (no_items)
	{
		int row = 0;
		foreach (QString s, vars)
		{
			QTableWidgetItem* item
					= new QTableWidgetItem(QString::number(0, 'f', VARS_PRECISION));
			m_variationsTable->setItem(row++, 0, item);
		}
		no_items = false;
	}

	flam3_xform* xform = selectedTriangle->xform();
	double* var = xform->var;
	int row = 0;
	foreach (QString s, vars)
	{
		QTableWidgetItem* item = m_variationsTable->item(row++, 0);
		double value = var[Util::variation_number(s)];
		item->setText(QString::number(value, 'f', VARS_PRECISION));
		m_variationsTable->colorItem(item, value);
	}
	m_variationsTable->blockSignals(false);


	m_variablesTable->blockSignals(true);
 	static bool no_items_2 = true;
	if (no_items_2)
	{
		int num_vars = Util::get_variable_names().size();
		for ( int row = 0; row < num_vars ; row++)
		{
			QTableWidgetItem* item
					= new QTableWidgetItem(QString::number(0, 'f', VARS_PRECISION));
			m_variablesTable->setItem(row, 0, item);
		}
		no_items_2 = false;
	}


	QStringList var_names = Util::get_variable_names();
	row = 0;
	foreach (QString s, var_names)
	{
		QTableWidgetItem* item = m_variablesTable->item(row++, 0);
		double value = Util::get_xform_variable(xform, s);
		item->setText(QString::number(value, 'f', VARS_PRECISION));
		m_variablesTable->colorItem(item, value);
	}
	m_variablesTable->blockSignals(false);

	updateLabel();
}

void VariationsWidget::variationEditedSlot(int row, int column)
{
	flam3_xform* xform = selectedTriangle->xform();
	double* var = xform->var;
	QString varName = m_variationsTable->verticalHeaderItem(row)->text();
	QTableWidgetItem* item = m_variationsTable->item(row,0);
	double value = QString(item->text()).toDouble();
	m_variationsTable->colorItem(item, value);
	var[Util::variation_number(varName)] = value;
	updateLabel();
	emit dataChanged();
}

void VariationsWidget::variableEditedSlot(int row, int column)
{
	flam3_xform* xform = selectedTriangle->xform();
	QString varName = m_variablesTable->verticalHeaderItem(row)->text();
	QTableWidgetItem* item = m_variablesTable->item(row,0);
	double value = QString(item->text()).toDouble();
	m_variablesTable->colorItem(item, value);
	Util::set_xform_variable(xform, varName, value);
	updateLabel();
	emit dataChanged();
}


void VariationsWidget::triangleSelectedSlot(Triangle* t)
{
	selectedTriangle = t;
	updateFormData();
}

/**
 * For some reason the wheelEvent occuring over the table headers is handled
 * here.  Throwing it down to the handler for the tablewidget causes an
 * infinite loop.
 */
void VariationsWidget::wheelEvent(QWheelEvent* e)
{
	QAbstractSlider::SliderAction a = QAbstractSlider::SliderSingleStepSub;
	if (e->delta() < 0)
		if (e->modifiers() & Qt::ControlModifier)
			a = QAbstractSlider::SliderPageStepAdd;
		else
			a = QAbstractSlider::SliderSingleStepAdd;
	else
		if (e->modifiers() & Qt::ControlModifier)
			a = QAbstractSlider::SliderPageStepSub;

	if (m_variationsTable->isVisible())
		m_variationsTable->verticalScrollBar()->triggerAction(a);
	else
		m_variablesTable->verticalScrollBar()->triggerAction(a);

	e->accept();
}



void VariationsWidget::updateLabel()
{
	int rows = m_variationsTable->rowCount();
	QStringList sl;
	for (int n = 0 ; n < rows ; n++)
	{

		QString val = m_variationsTable->item(n, 0)->text();
		if (val.toDouble() != 0.0)
		{
			QString var = m_variationsTable->verticalHeaderItem(n)->text();
			sl << QString("%1: %2").arg(var).arg(val);
		}
	}
	m_varListLabel->setWordWrap(true);
	m_varListLabel->setText(sl.join(QString(", ")));
}
