/***************************************************************************
 *   Copyright (C) 2007 by David Bitseff                                   *
 *   dbitsef@zipcon.net                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <QGraphicsTextItem>
#include <QGraphicsView>

#include "xfedit.h"
#include "logger.h"

#define genome_ptr (genome->selectedGenome())

/**
 * The QGraphicsScene that manages the Triangles.
 */
FigureEditor::FigureEditor(QObject* parent)
	: QGraphicsScene(parent), moving(0), selectedTriangle(0)
{
}


FigureEditor::FigureEditor(GenomeVector* g, QObject* parent)
	: QGraphicsScene(parent), moving(0), genome(g),
		selectedTriangle(0), postTriangle(0),
		has_selection(false), is_selecting(false), editing_post(false)
{
	QMatrix b(100.0, 0.0, 0.0, -100.0, 0.0, 0.0);
	basisTriangle = new BasisTriangle(b);

	popupMenu = new QMenu(tr("Edit Triangle"));
	addAction = new QAction(QIcon(":icons/silk/bullet_add.xpm"),
							tr("Add Triangle"),this);
	dupAction = new QAction(QIcon(":icons/silk/bullet_add.xpm"),
							tr("Duplicate Triangle"),this);
	delAction = new QAction(QIcon(":icons/silk/bullet_delete.xpm"),
							tr("Remove Triangle"),this);
	resAction = new QAction(QIcon(":icons/silk/shape_square_go.xpm"),
							tr("Reset Position"),this);
	flipHAction = new QAction(QIcon(":icons/silk/shape_flip_horizontal.xpm"),
							  tr("Flip Horizontal"),this);
	flipVAction = new QAction(QIcon(":icons/silk/shape_flip_vertical.xpm"),
							  tr("Flip Vertical"),this);
	rescaleAction = new QAction(QIcon(":icons/silk/shape_group.xpm"),
							  tr("Rescale"),this);

	triangleMenu = popupMenu->addMenu("Triangles");
	popupMenu->addSeparator();
	popupMenu->addAction(addAction);
	popupMenu->addAction(dupAction);
	popupMenu->addAction(delAction);
	popupMenu->addSeparator();
	popupMenu->addAction(flipHAction);
	popupMenu->addAction(flipVAction);
	popupMenu->addAction(resAction);
	popupMenu->addSeparator();
	popupMenu->addAction(rescaleAction);

	grid_visible = false;
	grid_color   = Qt::gray;
	bg_color     = Qt::black;

	setBackgroundBrush(Qt::black);
	basisTriangle->setGraphicsScene(this);
	addItem(basisTriangle);

	infoItem = new QGraphicsSimpleTextItem();
	addItem(infoItem);

	selectionItem = new TriangleSubset(this, basisTriangle);
	selectionItem->setPen(QPen(QBrush(Qt::gray),1,Qt::DashLine));
	selectionItem->setZValue(-1);
	selectionItem->setVisible(false);
	addItem(selectionItem);

	bLabelA = new QGraphicsSimpleTextItem(QString("O"));
	bLabelA->setPos(basisTriangle->polygon().at(0) + QPointF(-10.0,0.0));
	bLabelA->setBrush(QBrush(Qt::gray));
	bLabelA->setZValue(0);
	addItem(bLabelA);

	bLabelB = new QGraphicsSimpleTextItem(QString("X"));
	bLabelB->setPos(basisTriangle->polygon().at(0) + QPointF(-10.0,0.0));
	bLabelB->setBrush(QBrush(Qt::gray));
	bLabelB->setZValue(0);
	addItem(bLabelB);

	bLabelC = new QGraphicsSimpleTextItem(QString("Y"));
	bLabelC->setPos(basisTriangle->polygon().at(0) + QPointF(-10.0,0.0));
	bLabelC->setBrush(QBrush(Qt::gray));
	bLabelC->setZValue(0);
	addItem(bLabelC);

	coordinateMark = new CoordinateMark();
	coordinateMark->setPen(QPen(Qt::gray));
	addItem(coordinateMark);
	coordinateMark->centerOn(QPointF(0.0,0.0));
	coordinateMark->setVisible(false);
	coordinateMark->setZValue(0);

	connect(triangleMenu, SIGNAL(triggered(QAction*)),
			this, SLOT(triangleMenuAction(QAction*)));
	
	connect(addAction, SIGNAL(triggered()),
			this, SLOT(addTriangleAction()));
	
	connect(dupAction, SIGNAL(triggered()),
			this, SLOT(duplicateTriangleAction()));
	
	connect(delAction, SIGNAL(triggered()),
			this, SLOT(removeTriangleAction()));
	
	connect(resAction, SIGNAL(triggered()), 
			this, SLOT(resetTriangleCoordsAction()));
	
	connect(flipHAction, SIGNAL(triggered()),
			this, SLOT(flipTriangleHAction()));

	connect(flipVAction, SIGNAL(triggered()),
			this, SLOT(flipTriangleVAction()));

	connect(rescaleAction, SIGNAL(triggered()), this, SLOT(autoScale()));
}

FigureEditor::~FigureEditor()
{
	delete selectionItem;
	delete postTriangle;
	delete infoItem;
	delete bLabelA;
	delete bLabelB;
	delete bLabelC;
	delete coordinateMark;
	delete popupMenu;
	delete triangleMenu;
	delete addAction;
	delete dupAction;
	delete delAction;
	delete resAction;
	delete flipHAction;
	delete flipVAction;
	delete rescaleAction;
	delete basisTriangle;
}

void FigureEditor::addTriangleAction()
{
	int n = genome_ptr->num_xforms;
	Util::add_default_xform(genome_ptr);
	logFine(QString("FigureEditor::addTriangleAction : adding triangle to slot %1")
			.arg(n));
	reset();
	logFine(QString("FigureEditor::addTriangleAction : %1 items in triangleList")
			.arg(trianglesList.size()));
	logFine(QString("FigureEditor::addTriangleAction : genome has %1 xforms")
			.arg( genome_ptr->num_xforms ));
	selectTriangle(trianglesList.last());
	emit triangleModifiedSignal(selectedTriangle);
}

void FigureEditor::duplicateTriangleAction()
{
	Triangle* t = getCurrentOrSelected();
	// flam3_add_xforms() reallocs the xforms, so save the index
	// of triangle t for later instead of a ref
	int idx = trianglesList.indexOf(t, 0);
	int n = genome_ptr->num_xforms;
	flam3_add_xforms(genome_ptr, 1, 0);
	genome_ptr->xform[n] = genome_ptr->xform[idx];
	reset();
	selectTriangle(trianglesList.last());
	emit triangleModifiedSignal(selectedTriangle);
}

void FigureEditor::removeTriangleAction()
{
	Triangle* t = getCurrentOrSelected();
	int idx = trianglesList.indexOf(t, 0);
	logFine(QString("Removing triangle %1").arg(idx));
	flam3_delete_xform(genome_ptr, idx);
	reset();
	logFine(QString("there are %1 items in triangleList")
			.arg(trianglesList.size()));
	logFine(QString("genome has %1 xforms").arg( genome_ptr->num_xforms ));
	if (selectedTriangle)
	{// notify of scene change
		emit triangleModifiedSignal(selectedTriangle);
	}
}

Triangle* FigureEditor::getCurrentOrSelected()
{
	Triangle* t;
	QGraphicsItem* item = itemAt(moving_start);
	if (hasFocus() && item && item->type() == Triangle::RTTI)
		t = dynamic_cast<Triangle*>(item);
	else
		t = selectedTriangle;

	return t;
}

void FigureEditor::resetTriangleCoordsAction()
{
	Triangle* t;
	QGraphicsItem* item = itemAt(moving_start);
	if (hasFocus() && item)
		switch (item->type())
		{
			case Triangle::RTTI:
			case PostTriangle::RTTI:
				t = dynamic_cast<Triangle*>(item);
				break;
			default:
				;
		}
	else
		t = selectedTriangle;

	t->resetPosition();
	emit triangleModifiedSignal(selectedTriangle);
}

void FigureEditor::mousePressEvent(QGraphicsSceneMouseEvent* e)
{
	if (e->button() == Qt::RightButton)
	{
		moving_start = e->scenePos();
		popupMenu->popup(e->screenPos());
	}
	else if (e->button() == Qt::MidButton)
	{
		if (coordinateMark->isVisible())
		{
			coordinateMark->centerOn(e->scenePos());
			if (!trianglesList.isEmpty())
				coordinateMark->setZValue(trianglesList.first()->nextZPos());
		}
	}
	else if (e->modifiers() & Qt::ControlModifier) // LeftButton+Ctrl to select
	{
		is_selecting = true;
		moving_start = e->scenePos();
		selectionItem->setPos(QPointF(0.0,0.0));
		selectionItem->setPolygon(QPolygonF(QRectF(moving_start,QSizeF(0.01,0.01))));
		selectionItem->setVisible(true);
		selectionItem->clear();
	}
	else // LeftButton, no modifier
	{
		QGraphicsItem* item = itemAt(e->scenePos());
		if (item == infoItem) // skip the infoItem if possible
		{
			QList<QGraphicsItem*> list = items(e->scenePos());
			if (list.size() > 1)
				item = list.at(1);
		}
		if (item)
		{
			moving = 0;
			logFine(QString("FigureEditor::mousePressEvent : rtti %1")
					.arg(item->type()));
			logFine(QString("FigureEditor::mousePressEvent : Found item %1")
					.arg((long)item));

			if ( item->type() == Triangle::RTTI )
			{
				Triangle* t = dynamic_cast<Triangle*>(item);
				moving = t;
				selectTriangle(t);
			}
			else if ( item->type() == PostTriangle::RTTI )
			{
				PostTriangle* t = dynamic_cast<PostTriangle*>(item);
				moving = t;
			}
			else if ( item->type() == NodeItem::RTTI )
			{
				NodeItem* node = dynamic_cast<NodeItem*>(item);
				Triangle* t = node->triangle();
				node->setZValue(t->nextZPos());
				moving = node;
				if (t->type() == Triangle::RTTI)
					selectTriangle(t);
			}
			else if ( item->type() == TriangleSubset::RTTI )
			{
				moving = dynamic_cast<TriangleSubset*>(item);
				if (!trianglesList.isEmpty())
					moving->setZValue(trianglesList.first()->nextZPos());
			}
			else
				return;

			moving_start = e->scenePos();
		}
		else
			logFine("FigureEditor::mousePressEvent : no item selected");
	}
}

void FigureEditor::mouseReleaseEvent(QGraphicsSceneMouseEvent* e)
{
    if (moving)
			moving = 0;

	else if (is_selecting)
	{
		is_selecting = false;
		if (e->scenePos() == e->buttonDownScenePos(Qt::LeftButton))
		{
			selectionItem->setPolygon(QPolygonF(QRectF(QPointF(0.0,0.0),QSizeF(0.1,0.1))));
			selectionItem->setPos(QPointF(0.0,0.0));
			selectionItem->setVisible(false);
			selectionItem->setZValue(-1);
			selectionItem->clear();
			has_selection = false;
		}
		else
		{
			QList<QGraphicsItem*> collisions = selectionItem->collidingItems();
			foreach (QGraphicsItem* item, collisions)
				if (item->type() == Triangle::RTTI)
				{
					logFine(QString("FigureEditor::mouseReleaseEvent : adding triangle %1,%2")
							.arg(item->scenePos().x())
							.arg(item->scenePos().y()));
					selectionItem->append(dynamic_cast<Triangle*>(item));
				}
			selectionItem->setZValue(selectedTriangle->nextZPos());
			has_selection = true;
		}
	}
}

void FigureEditor::mouseMoveEvent(QGraphicsSceneMouseEvent* e)
{
	static Triangle* lastt;
	QPointF p = e->scenePos();
	QPointF s = e->screenPos();

	// I guess we're only interested in scenePos() changes?  There
	// is an undocumented feature where this event handler is called
	// when the sceneRect is changed.
	if (s == e->lastScreenPos())
		return;

	QGraphicsItem* item = itemAt(p);

	if ( moving )
	{
		infoItem->hide();
		lastt = 0;
		if ( moving->type() == Triangle::RTTI )
		{
			Triangle* t = dynamic_cast<Triangle*>(moving);
			logFiner(QString("FigureEditor::mouseMoveEvent : "
						"moving triangle by %1,%2")
					.arg(p.x() - moving_start.x())
					.arg(p.y() - moving_start.y()));
			t->moveBy(p.x() - moving_start.x(), p.y() - moving_start.y());
			emit triangleModifiedSignal(t);
		}
		else if ( moving->type() == TriangleSubset::RTTI )
		{
			selectionItem->moveBy(p.x() - moving_start.x(),
							p.y() - moving_start.y());
			logFiner(QString("FigureEditor::mouseMoveEvent : "
						"triangle set moved to %1,%2")
					.arg(selectionItem->scenePos().x())
					.arg(selectionItem->scenePos().y()));
			if (selectionItem->size() > 0)
				emit triangleModifiedSignal(selectionItem->first());
		}
		else if ( moving->type() == PostTriangle::RTTI )
		{
			dynamic_cast<PostTriangle*>(moving)->moveBy(
				p.x() - moving_start.x(), p.y() - moving_start.y());
			emit triangleModifiedSignal(selectedTriangle);
		}
		else // NodeItem::RTTI
		{
			NodeItem* node = dynamic_cast<NodeItem*>(moving);
			node->moveBy(p.x() - moving_start.x(), p.y() - moving_start.y());
			emit triangleModifiedSignal(selectedTriangle);
		}
		moving_start = p;
		update();
    }
	else if (is_selecting)
		selectionItem->setPolygon(QPolygonF(QRectF(moving_start,p)));
	
	else if (coordinateMark->isVisible() && (e->buttons() & Qt::MidButton))
		coordinateMark->centerOn(e->scenePos());

	else if (item)
	{
		switch (item->type())
		{
			case NodeItem::RTTI:
				item = dynamic_cast<NodeItem*>(item)->triangle();

			case Triangle::RTTI:
				if (item->type() != PostTriangle::RTTI)
				{
					// show the xform's infoItem for this triangle
					Triangle* t = dynamic_cast<Triangle*>(item);
					if (lastt != t)
					{
						infoItem->setText(getInfoLabel(t));
						QPoint pt = QPoint(5,
								views().first()->maximumViewportSize().height()
								+ 10 - (int)infoItem->boundingRect().height());
								// 10 units of fudge
						infoItem->setPos(views().first()->mapToScene(pt));
						infoItem->setPen(t->pen());
						infoItem->setZValue(selectedTriangle->zValue());
						infoItem->show();
						lastt = t;
					}
					break;
				}
			case PostTriangle::RTTI:
			{
				Triangle* t = selectedTriangle;
				int tidx = trianglesList.indexOf(t);
				QString info = QString("post transform: %1\n").arg(tidx + 1);
				infoItem->setText(info);
				QPoint pt = QPoint(5,
						views().first()->maximumViewportSize().height()
						+ 10 - (int)infoItem->boundingRect().height());
						// 10 units of fudge
				infoItem->setPos(views().first()->mapToScene(pt));
				infoItem->setPen(t->pen());
				infoItem->setZValue(selectedTriangle->zValue());
				infoItem->show();
				break;
			}
			default:
				;
		}
	}
	else if (item != infoItem && infoItem->isVisible())
	{
		infoItem->hide();
		infoItem->setZValue(0);
		infoItem->setText(QString("0"));
		infoItem->setPos(QPointF(0.0,0.0));
		lastt = 0;
	}

	emit coordinateChangeSignal( p.x(), p.y() );
}


void FigureEditor::wheelEvent(QGraphicsSceneWheelEvent* e)
{
	QPointF p = e->scenePos();
	QGraphicsItem* item = itemAt(p);
	if (item)
	{
		switch (item->type())
		{
			case Triangle::RTTI:
			case PostTriangle::RTTI:
			{
		
				double rad;
				Triangle* t;
				if (item->type() == Triangle::RTTI)
					t = dynamic_cast<Triangle*>(item);
				else
					t = dynamic_cast<PostTriangle*>(item);

				QPointF cpos = t->mapFromScene(p);
				if (e->modifiers() & Qt::ControlModifier)
				{
					// scale
					if (e->delta() > 0)
						if (e->modifiers() & Qt::ShiftModifier)
							rad = 1.05;
						else
							rad = 1.1;
					else
						if (e->modifiers() & Qt::ShiftModifier)
							rad = 0.9523;
						else
							rad = 0.9090;
					scaleTriangle(rad, rad, cpos, t);
				}
				else
				{
					// rotate
					if (e->modifiers() & Qt::ShiftModifier)
						rad = (double)e->delta() / (128.*3.141592654);
					else
						rad = (double)e->delta() / (16.*3.141592654);
					rotateTriangle(rad, cpos, t);
				}
				e->accept();
				break;
			}
			case BasisTriangle::RTTI:
			{
				double rad;
				if (e->delta() > 0)
					if (e->modifiers() & Qt::ShiftModifier)
						rad = 1.1;
					else
						rad = 1.2;
				else
					if (e->modifiers() & Qt::ShiftModifier)
						rad = 0.9090;
					else
						rad = 0.8333;
				scaleBasis(rad,rad);
				e->accept();
				break;
			}
			case TriangleSubset::RTTI:
			{
				double rad;
				QPointF cpos = selectionItem->mapFromScene(p);
				if (e->modifiers() & Qt::ControlModifier)
				{
					// scale
					if (e->delta() > 0)
						if (e->modifiers() & Qt::ShiftModifier)
							rad = 1.05;
						else
							rad = 1.1;
					else
						if (e->modifiers() & Qt::ShiftModifier)
							rad = 0.9523;
						else
							rad = 0.9090;
					scaleSelection(rad, rad, cpos);
				}
				else
				{
					// rotate
					if (e->modifiers() & Qt::ShiftModifier)
						rad = (double)e->delta() / (128.*3.141592654);
					else
						rad = (double)e->delta() / (16.*3.141592654);
					rotateSelection(rad, cpos);
				}
				e->accept();
			}
			default:
			;
		}
	}
}

void FigureEditor::keyPressEvent(QKeyEvent* e)
{
	switch(e->key())
	{
		case Qt::Key_Semicolon:
		case Qt::Key_PageUp:
			if (e->modifiers() & Qt::ControlModifier)
				scaleBasis(2., 2.);
			else
				scaleBasis(1.1, 1.1);
			e->accept();
			break;

		case Qt::Key_Colon:
		case Qt::Key_PageDown:
			if (e->modifiers() & Qt::ControlModifier)
				scaleBasis(.5, .5);
			else
				scaleBasis(.9090, .9090);
			e->accept();
			break;
		default:
			QGraphicsScene::keyPressEvent(e);
	}
}

void FigureEditor::triangleModifiedAction(Triangle* t)
{
	logFinest(QString("FigureEditor::triangleModifiedAction : t=0x%1")
			.arg((long)t, 0, 16));
	selectTriangle(t);
	update();
	emit triangleModifiedSignal(t);
}

void FigureEditor::flipTriangleHAction()
{
	if (has_selection)
		flipSelectionHAction(selectionItem->mapFromScene(moving_start));
	else if (editing_post)
		flipTriangleHAction(postTriangle->mapFromScene(moving_start));
	else
		flipTriangleHAction(selectedTriangle->mapFromScene(moving_start));
}

void FigureEditor::flipTriangleHAction(QPointF cpos)
{
	if (editing_post)
		postTriangle->flipHorizontally(cpos);
	else
		selectedTriangle->flipHorizontally(cpos);
	update();
	emit triangleModifiedSignal(selectedTriangle);
}

void FigureEditor::flipTriangleVAction()
{
	if (has_selection)
		flipSelectionVAction(selectionItem->mapFromScene(moving_start));
	else if (editing_post)
		flipTriangleVAction(postTriangle->mapFromScene(moving_start));
	else
		flipTriangleVAction(selectedTriangle->mapFromScene(moving_start));
}

void FigureEditor::flipTriangleVAction(QPointF cpos)
{
	if (editing_post)
		postTriangle->flipVertically(cpos);
	else
		selectedTriangle->flipVertically(cpos);
	update();
	emit triangleModifiedSignal(selectedTriangle);
}

void FigureEditor::flipSelectionVAction(QPointF cpos)
{
	selectionItem->flipVertically(cpos);
	if (selectionItem->size() > 0)
		triangleModifiedAction(selectionItem->first());
}

void FigureEditor::flipSelectionHAction(QPointF cpos)
{
	selectionItem->flipHorizontally(cpos);
	if (selectionItem->size() > 0)
		triangleModifiedAction(selectionItem->first());
}

BasisTriangle* FigureEditor::basis() const
{
	return basisTriangle;
}

void FigureEditor::setCoordTransform(QMatrix m)
{
	basisTriangle->setCoordTransform(m);
}

const QMatrix& FigureEditor::transform() const
{
	return basisTriangle->coordTransform();
}

void FigureEditor::selectTriangle(Triangle* t)
{
	if (selectedTriangle == t)
		return;

	bool post = editing_post;

	if (selectedTriangle)
	{
		selectedTriangle->setPen(
			QPen(Util::get_xform_color(genome_ptr,selectedTriangle->xform())));
		selectedTriangle->setBrush(Qt::NoBrush);
		triangleMenu->actions().value(selectedTriangleIndex())->setChecked(false);
		editPostTriangle(false);
	}

	selectedTriangle = t;
	logFine(QString("FigureEditor::selectTriangle : selecting triangle %1")
			.arg(trianglesList.indexOf(t, 0) + 1));
	const QColor c = Util::get_xform_color(genome_ptr, selectedTriangle->xform());
	selectedTriangle->setPen(QPen(c));
	selectedTriangle->setBrush(QBrush(c, Qt::Dense4Pattern));
	selectedTriangle->moveToFront();
	triangleMenu->actions().value(selectedTriangleIndex())->setChecked(true);
	editPostTriangle(post);
	emit triangleSelectedSignal(t);
}

void FigureEditor::reset()
{
	logFine("FigureEditor::reset : entered");
	int selected_idx = 0;
	if (!trianglesList.isEmpty())
	{
		if (selectedTriangle)
			selected_idx = selectedTriangleIndex();

		QList<Triangle*>::iterator i;
		for (i = trianglesList.begin(); i != trianglesList.end(); ++i)
		{
			removeItem(*i);
			delete *i;
		}
		trianglesList.clear();
		triangleMenu->clear();
	}
	bool hasPost = editing_post;
	editPostTriangle(false);

	flam3_xform* xform = genome_ptr->xform;
	for (int n = 0 ; n < genome_ptr->num_xforms ; n++)
	{
		Triangle* t = new Triangle(this, xform + n, basisTriangle);
		const QPen pen( Util::get_xform_color(genome_ptr, xform + n ) );
		t->setPen( pen );
		t->setNodeColor(grid_color, pen.color());
		trianglesList << t;
		addItem(t);
		QString action_name = QString("%1").arg(n+1);
		QAction* a = triangleMenu->addAction(action_name);
		a->setIcon(QIcon(":icons/silk/bullet_black.xpm"));
		a->setObjectName(action_name);
		a->setCheckable(true);
	}
	const QBrush b(grid_color);
	bLabelA->setBrush(b);
	bLabelA->setPos(basisTriangle->mapToScene(basisTriangle->A)
			+ QPointF(-8.0,0.0));
	bLabelB->setBrush(b);
	bLabelB->setPos(basisTriangle->mapToScene(basisTriangle->B));
	bLabelC->setBrush(b);
	bLabelC->setPos(basisTriangle->mapToScene(basisTriangle->C)
			+ QPointF(-8.0,-16.0));
	const QPen pen(grid_color);
	basisTriangle->setPen(pen);
	coordinateMark->setPen(pen);
	selectionItem->setPen(QPen(b,1,Qt::DashLine));

	selectedTriangle = 0;
	editing_post = hasPost; // there's a side-effect here that selectTriangle
							// will recreate the postTriangle with this set
	if (!trianglesList.isEmpty()) // try to select the same triangle
	{
		if (selected_idx < trianglesList.size())
			selectTriangle(trianglesList.at(selected_idx));
		else
			selectTriangle(trianglesList.last());
	}

	// Adjust the sceneRect so it covers just more than the items.  This
	// helps reposition the sceneview and adjusts the scrollbars.  It also
	// helps avoid some "jumpiness" that sometimes happens when moving
	// a graphicsitem outside of the sceneRect.
	if (infoItem->isVisible())
	{
		// hide/move the infoItem first so it doesn't enlarge the boundingRect
		removeItem(infoItem);
		setSceneRect(itemsBoundingRect().adjusted(-200.,-200.,200.,200.));
		// put the infoItem back
		addItem(infoItem);
		QPoint pt = QPoint(5,
			views().first()->maximumViewportSize().height()
				+ 10 - (int)infoItem->boundingRect().height());
				// 10 units of fudge
		infoItem->setPos(views().first()->mapToScene(pt));
	}
	else
		setSceneRect(itemsBoundingRect().adjusted(-200.,-200.,200.,200.));

	logFiner(QString("FigureEditor::reset : sceneRect %1,%2")
			.arg(sceneRect().width())
			.arg(sceneRect().height()));
	logFiner(QString("FigureEditor::reset : itemsRect %1,%2")
			.arg(itemsBoundingRect().size().width())
			.arg(itemsBoundingRect().size().height()));
	logFiner(QString("FigureEditor::reset : viewport %1,%2")
			.arg(views().first()->maximumViewportSize().width())
			.arg(views().first()->maximumViewportSize().height()));

	if (has_selection)
	{
		selectionItem->setPolygon(QPolygonF(QRectF(QPointF(0.0,0.0),QSizeF(0.1,0.1))));
		selectionItem->setPos(QPointF(0.0,0.0));
		selectionItem->setVisible(false);
		selectionItem->setZValue(-1);
		selectionItem->clear();
		has_selection = false;
	}
}

void FigureEditor::drawBackground(QPainter* p, const QRectF& r)
{
	QRectF rect(sceneRect().united(r));
	p->fillRect(r, QBrush(bg_color));
	if (grid_visible)
	{
		// draw a grid on the background
		QPointF A = basisTriangle->mapToScene(basisTriangle->A);
		QPointF B = basisTriangle->mapToScene(basisTriangle->B);
		QPointF C = basisTriangle->mapToScene(basisTriangle->C);
		QPointF dX = B - A;
		QPen pen(QBrush(grid_color, Qt::Dense6Pattern),Qt::DashLine);
		p->setPen(pen);
		for (qreal n = A.x() ; n <= rect.width() ; n+=dX.x())
			p->drawLine(QPointF(n, rect.y()), QPointF(n, rect.height()));
		for (qreal n = A.x() - dX.x() ; n >= rect.x() ; n-=dX.x())
			p->drawLine(QPointF(n, rect.y()), QPointF(n, rect.height()));
		dX = B - C;
		for (qreal n = C.y() ; n <= rect.height() ; n += dX.y())
			p->drawLine(QPointF(rect.x(), n), QPointF(rect.width(), n));
		for (qreal n = C.y() - dX.y() ; n >= rect.y() ; n -= dX.y())
			p->drawLine(QPointF(rect.x(), n), QPointF(rect.width(), n));
	}
}

void FigureEditor::colorChangedAction(double idx)
{
	logFinest(QString("FigureEditor::colorChangedAction : idx=0x%1").arg(idx));
	QList<Triangle*>::iterator i;
	for (i = trianglesList.begin(); i != trianglesList.end(); ++i)
	{
		QColor c = Util::get_xform_color(genome_ptr, (*i)->xform());
		const QPen pen(c);
		(*i)->setPen(pen);
		if (*i == selectedTriangle)
			(*i)->setBrush(QBrush(c, Qt::Dense4Pattern));
		else
			(*i)->setBrush(QBrush(c, Qt::NoBrush));
		foreach (NodeItem* n, (*i)->getNodes())
			n->setPen(QPen(grid_color), c);
	}
}

void FigureEditor::setSceneRect (const QRectF& rect)
{
	QGraphicsScene::setSceneRect(rect);
	emit sceneRectChanged(rect);
}

int FigureEditor::selectedTriangleIndex()
{
	return trianglesList.indexOf(selectedTriangle);
}

Triangle* FigureEditor::getSelectedTriangle()
{
	return selectedTriangle;
}

int FigureEditor::getNumberOfTriangles()
{
	return trianglesList.size();
}

bool FigureEditor::gridVisible() const
{
	return grid_visible;
}

void FigureEditor::setGridVisible(bool value)
{
	grid_visible = value;
	update();
}


QColor FigureEditor::gridColor() const
{
	return grid_color;
}

void FigureEditor::setGridColor(const QColor& c)
{
	grid_color = c;
	const QPen p(grid_color);
	const QBrush b(grid_color);
	basisTriangle->setPen(p);
	coordinateMark->setPen(p);
	bLabelA->setBrush(b);
	bLabelB->setBrush(b);
	bLabelC->setBrush(b);
	selectionItem->setPen(QPen(b,1,Qt::DashLine));
	foreach(Triangle* t, trianglesList)
		t->setNodeColor(c, t->pen().color());
	update();
}

QColor FigureEditor::bgColor() const
{
	return bg_color;
}

void FigureEditor::setbgColor(QColor c)
{
	bg_color = c;
	update();
}

void FigureEditor::triangleMenuAction(QAction* a)
{
	int idx = a->objectName().toInt() - 1;
	selectTriangle(idx);
}

void FigureEditor::selectTriangle(int idx)
{
	if (idx < trianglesList.size())
	{
		Triangle* t = trianglesList.at(idx);
		selectTriangle(t);
		QList<QGraphicsView*> viewList = views();
		viewList.first()->ensureVisible(t->sceneBoundingRect());
	}
}

CoordinateMark* FigureEditor::mark() const
{
	return coordinateMark;
}

void FigureEditor::setMarkVisible(bool flag)
{
	coordinateMark->setVisible(flag);
	if (flag && !trianglesList.isEmpty())
		coordinateMark->setZValue(trianglesList.first()->nextZPos());
	emit coordinateChangeSignal(0.,0.);
}

void FigureEditor::autoScale()
{
	// hide/move the infoItem first so it doesn't enlarge the boundingRect
	if (infoItem->isVisible())
	{
		infoItem->hide();
		infoItem->setZValue(0);
		infoItem->setText(QString("0"));
		infoItem->setPos(QPointF(0.0,0.0));
	}
	QSizeF s_size = itemsBoundingRect().size() + QSizeF(100.,100.);
	QSize v_size = views().first()->maximumViewportSize();
	double dz = qMin(v_size.width() / s_size.width(),
				v_size.height() / s_size.height());
	while (dz < 0.99 || dz > 1.01)
	{
		logFine(QString("FigureEditor::autoScale : scale %1").arg(dz));
		scaleBasis(dz, dz);
		s_size = itemsBoundingRect().size() + QSizeF(100.,100.);
		v_size = views().first()->maximumViewportSize();
		dz = qMin(v_size.width() / s_size.width(),
				v_size.height() / s_size.height());
	}
	views().first()->ensureVisible(itemsBoundingRect(),0,0);
}

void FigureEditor::selectNextTriangle()
{
	int n = selectedTriangleIndex();
	if (n + 1 < trianglesList.size())
		selectTriangle(n + 1);
}

void FigureEditor::selectPreviousTriangle()
{
	int n = selectedTriangleIndex();
	if (n > 0)
		selectTriangle(n - 1);
}

void FigureEditor::scaleBasis(double dx, double dy)
{
	basisTriangle->scale(dx, dy);
	reset();
}

void FigureEditor::rotateTriangle(double rad, QPointF pos, Triangle* t)
{
	if (editing_post)
	{
		postTriangle->rotate(rad, pos);
		triangleModifiedAction(selectedTriangle);
	}
	else
	{
		t->rotate(rad, pos);
		triangleModifiedAction(t);
	}
}

void FigureEditor::scaleTriangle( double dx, double dy, QPointF pos, Triangle* t)
{
	if (editing_post)
	{
		postTriangle->scale(dx, dy, pos);
		triangleModifiedAction(selectedTriangle);
	}
	else
	{
		t->scale(dx, dy, pos);
		triangleModifiedAction(t);
	}
}

void FigureEditor::rotateSelection( double rad, QPointF pos )
{
	selectionItem->rotate(rad, pos);
	logFine(QString("triangle set rotated rad %1 (%2,%3)")
				.arg(rad)
				.arg(pos.x())
				.arg(pos.y()));
	if (selectionItem->size() > 0)
		triangleModifiedAction(selectionItem->first());
}

void FigureEditor::scaleSelection( double dx, double dy, QPointF pos )
{
	selectionItem->scale(dx, dy, pos);
	logFine(QString("triangle set scaled dx %1 (%2,%3)")
				.arg(dx)
				.arg(pos.x())
				.arg(pos.y()));
	if (selectionItem->size() > 0)
		triangleModifiedAction(selectionItem->first());
}

bool FigureEditor::hasSelection() const
{
	return has_selection;
}


TriangleSubset* FigureEditor::selection() const
{
	return selectionItem;
}


void FigureEditor::moveTriangleBy(double dx, double dy, Triangle* t)
{
	t->moveTransformBy(dx, dy);
	if (t->type() == PostTriangle::RTTI)
		triangleModifiedAction(selectedTriangle);
	else
		triangleModifiedAction(t);
}


void FigureEditor::moveSelectionBy(double dx, double dy)
{
	QPointF p = selectionItem->mapFromItem(selectedTriangle, QPointF(dx,dy));
	selectionItem->moveBy(p.x(), p.y());
	logFine(QString("triangle set moved %1,%2")
			.arg(selectionItem->scenePos().x())
			.arg(selectionItem->scenePos().y()));
	if (selectionItem->size() > 0)
		triangleModifiedAction(selectionItem->first());
}


void FigureEditor::editPostTriangle(bool flag)
{
	if (flag) // enable post triangle
	{
		if (postTriangle)
		{
			removeItem(postTriangle);
			delete postTriangle;
		}

		flam3_xform* xf = selectedTriangle->xform();
		postTriangle = new PostTriangle(this, xf, basisTriangle);
		const QPen pen( grid_color );
		postTriangle->setPen( pen );
		postTriangle->setNodeColor(grid_color, pen.color());
		postTriangle->setVisible(true);
		postTriangle->setBrush(QBrush(grid_color, Qt::Dense4Pattern));
		addItem(postTriangle);
		postTriangle->adjustSceneRect();
	}
	else
	{
		if (postTriangle)
		{
			removeItem(postTriangle);
			delete postTriangle;
			postTriangle = 0;
		}
	}
	editing_post = flag;
}

bool FigureEditor::postEnabled()
{
	return editing_post;
}

PostTriangle* FigureEditor::post() const
{
	return postTriangle;
}

QString FigureEditor::getInfoLabel(Triangle* t)
{
	QString xfinal("");
	int tidx = trianglesList.indexOf(t);
	if (genome_ptr->final_xform_enable
			&& genome_ptr->final_xform_index == tidx)
		xfinal = "final ";

	QString info = QString("%1transform: %2\ndensity: %3\n")
					.arg(xfinal)
					.arg(tidx + 1)
					.arg(t->xform()->density);
	QMap<QString, int> vars = Util::flam3_variations();
	QMap<QString, int>::const_iterator i = vars.begin();
	while (i != vars.end())
	{
		double val = t->xform()->var[i.value()];
		if (val != 0.0)
			info.append(QString("%1: %2\n")
					.arg(i.key()).arg(val));
		i++;
	}
	return info;
}


const TriangleList& FigureEditor::triangles() const
{
	return trianglesList;
}

